from __future__ import annotations

from typing import Literal

import pytest
from htmltools import Tag, TagAttrValue

from shiny import ui
from shiny.ui._sidebar import SidebarOpenSpec, SidebarOpenValue


@pytest.mark.parametrize(
    "open_value, expected",
    [
        ("closed", {"desktop": "closed", "mobile": "closed"}),
        ("open", {"desktop": "open", "mobile": "open"}),
        ("always", {"desktop": "always", "mobile": "always"}),
        ("desktop", {"desktop": "open", "mobile": "closed"}),
    ],
)
def test_sidebar_open_string_values(
    open_value: SidebarOpenValue, expected: SidebarOpenSpec
):
    assert ui.sidebar(open=open_value).open() == ui.sidebar(open=expected).open()


def get_sidebar_tags(sb: ui.Sidebar) -> tuple[Tag, Tag]:
    sidebar, collapse = sb.tagify()
    assert isinstance(sidebar, Tag)
    assert isinstance(collapse, Tag)
    return sidebar, collapse


def test_sidebar_assigns_input_binding_class_if_id_provided():
    sidebar_tag, _ = get_sidebar_tags(ui.sidebar(id="my_sidebar"))

    assert sidebar_tag.has_class("bslib-sidebar-input")
    assert sidebar_tag.attrs["id"] == "my_sidebar"


def test_sidebar_assigns_random_id_if_collapsible_and_id_not_provided():
    s_open_sb, s_open_collapse = get_sidebar_tags(ui.sidebar(open="open"))

    assert s_open_sb.attrs["id"].startswith("bslib_sidebar_")
    assert s_open_sb.attrs["id"] == s_open_collapse.attrs["aria-controls"]

    s_closed_sb, s_closed_collapse = get_sidebar_tags(ui.sidebar(open="closed"))
    assert s_closed_sb.attrs["id"].startswith("bslib_sidebar_")
    assert s_closed_sb.attrs["id"] == s_closed_collapse.attrs["aria-controls"]

    s_always_sb, s_always_collapse = get_sidebar_tags(ui.sidebar(open="always"))
    assert "id" not in s_always_sb.attrs
    assert "aria-controls" not in s_always_collapse.attrs


def test_sidebar_sets_aria_expanded_on_collapse_toggle():
    def get_sidebar_collapse_aria_expanded(
        open: SidebarOpenValue | Literal["desktop"],
    ) -> TagAttrValue:
        _, collapse_tag = get_sidebar_tags(ui.sidebar(open=open))
        return collapse_tag.attrs["aria-expanded"]

    assert get_sidebar_collapse_aria_expanded("open") == "true"
    assert get_sidebar_collapse_aria_expanded("closed") == "false"
    assert get_sidebar_collapse_aria_expanded("desktop") == "true"

    _, collapse_always = get_sidebar_tags(ui.sidebar(open="always"))
    assert "aria-expanded" not in collapse_always.attrs


def test_sidebar_throws_for_invalid_open():
    with pytest.raises(ValueError, match="`open` must be a string matching"):
        ui.sidebar(open="bad")  # pyright: ignore[reportArgumentType]

    with pytest.raises(ValueError, match="`open` must be one of"):
        ui.sidebar(open=("closed", "open"))  # pyright: ignore[reportArgumentType]

    with pytest.raises(ValueError, match="`desktop` must be one of"):
        ui.sidebar(open={"desktop": "bad"})  # pyright: ignore[reportArgumentType]

    with pytest.raises(TypeError, match="widescreen"):
        ui.sidebar(open={"widescreen": "open"})  # pyright: ignore[reportArgumentType]



# # Within file flags to ignore unused imports
# flake8: noqa: F401
# pyright: reportUnusedImport=false

__all__ = (
    "Annotated",
    "Concatenate",
    "ParamSpec",
    "TypeGuard",
    "TypeIs",
    "Never",
    "Required",
    "NotRequired",
    "Self",
    "TypedDict",
    "assert_type",
)


import sys

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec, TypeGuard
else:
    from typing_extensions import Concatenate, ParamSpec, TypeGuard

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import Never, NotRequired, Required, Self, TypedDict, assert_type
else:
    from typing_extensions import (
        Never,
        NotRequired,
        Required,
        Self,
        TypedDict,
        assert_type,
    )

if sys.version_info >= (3, 13):
    from typing import TypeIs
else:
    from typing_extensions import TypeIs

# The only purpose of the following line is so that pyright will put all of the
# conditional imports into the .pyi file when generating type stubs. Without this line,
# pyright will not include the above imports in the generated .pyi file, and it will
# result in a lot of red squiggles in user code.
_: 'Annotated |Concatenate[str, ParamSpec("P")] | ParamSpec | TypeGuard | TypeIs | NotRequired | Required | TypedDict | assert_type | Self'  # type:ignore



from pathlib import Path

import pytest
from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc, run_shiny_app


@pytest.mark.parametrize(
    "app_path, sel_card, sel_vb",
    [
        ("app-express.py", "my_card", "my_value_box"),
        ("app-module.py", "card-thing", "value_box-thing"),
    ],
)
def test_card_input(page: Page, app_path: str, sel_card: str, sel_vb: str) -> None:
    sa: ShinyAppProc = run_shiny_app(
        Path(__file__).parent / app_path, wait_for_start=True
    )

    page.goto(sa.url)

    card = controller.Card(page, sel_card)
    vb = controller.ValueBox(page, sel_vb)
    out_card = controller.OutputCode(page, "out_card")
    out_vb = controller.OutputCode(page, "out_value_box")

    # Open and close card full screen, check input value ------
    card.expect_full_screen(False)
    out_card.expect_value("False")

    card.set_full_screen(True)
    card.expect_full_screen(True)
    out_card.expect_value("True")

    card.set_full_screen(False)
    card.expect_full_screen(False)
    out_card.expect_value("False")

    # Open and close value box full screen, check input value ------
    vb.expect_full_screen(False)
    out_vb.expect_value("False")

    vb.set_full_screen(True)
    vb.expect_full_screen(True)
    out_vb.expect_value("True")

    vb.set_full_screen(False)
    vb.expect_full_screen(False)
    out_vb.expect_value("False")



from __future__ import annotations

from playwright.sync_api import Locator, Page
from playwright.sync_api import expect as playwright_expect

from ...types import MISSING, MISSING_TYPE
from .._types import ListPatternOrStr, Timeout
from ..expect._expect import _attr_match_str, _xpath_match_str
from ._base import is_missing


def assert_arr_is_unique(
    arr: ListPatternOrStr,
    msg: str,
) -> None:
    """
    Assert that the array is unique.

    Parameters
    ----------
    arr
        The array to check.
    msg
        The error message.
    """
    assert len(arr) == len(list(dict.fromkeys(arr))), msg


def checked_css_str(
    is_checked: bool | MISSING_TYPE = MISSING,
) -> str:
    """
    Get the CSS string for checked elements.

    Parameters
    ----------
    is_checked
        Whether the elements are checked. Defaults to `MISSING`.
    """
    if is_missing(is_checked):
        return ""
    if is_checked:
        return ":checked"
    else:
        raise NotImplementedError("`is_checked = FALSE` is not verified yet")
        return ":not(:checked)"


def expect_locator_contains_values_in_list(
    *,
    page: Page,
    loc_container: Locator,
    el_type: str,
    arr_name: str,
    arr: list[str],
    is_checked: bool | MISSING_TYPE = MISSING,
    timeout: Timeout = None,
    key: str = "value",
) -> None:
    """
    Expect the locator to contain the values in the list.

    The matching values must exist and be in order, but other values may also exist
    within the container.

    Parameters
    ----------
    page
        Playwright `Page` of the Shiny app.
    loc_container
        The container locator.
    el_type
        The element type.
    arr_name
        The variable name.
    arr
        The expected values.
    is_checked
        Whether the elements are checked. Defaults to `MISSING`.
    timeout
        The timeout for the expectation. Defaults to `None`.
    key
        The key. Defaults to `"value"`.
    """
    # Make sure the locator contains all of `arr`
    if not isinstance(arr, list):
        raise TypeError(f"`{arr_name}` must be a list")
    for item in arr:
        if not isinstance(item, str):
            raise TypeError(f"`{arr_name}` must be a list of strings")

    # Make sure the locator has len(uniq_arr) input elements
    assert_arr_is_unique(arr, f"`{arr_name}` must be unique")
    is_checked_str = checked_css_str(is_checked)

    # If there are no items, then the container needs to exist.
    # All containers contain 0 items.
    if len(arr) == 0:
        playwright_expect(loc_container).to_have_count(1, timeout=timeout)
        return

    loc_container_orig = loc_container

    # Find all items in set
    for item in arr:
        # Given the container, make sure it contains this locator
        loc_container = loc_container.locator(
            "xpath=.",
            # Simple approach as position is not needed
            has=page.locator(
                f"{el_type}[{_attr_match_str(key, item)}]{is_checked_str}",
            ),
        )

    # If we are only looking to see if *some* (not *these only*) elements exist,
    # then we only need to check if the container locator (which must contain the elements) can be found
    try:
        playwright_expect(loc_container).to_have_count(1, timeout=timeout)
    except AssertionError as e:
        # Debug expections

        # Expecting container to exist (count = 1)
        playwright_expect(loc_container_orig).to_have_count(1, timeout=timeout)

        for item in arr:
            # Expecting item `{item}` to exist in container
            # Perform exact matches on strings.
            playwright_expect(
                # Simple approach as position is not needed
                loc_container_orig.locator(
                    f"{el_type}[{_attr_match_str(key, item)}]{is_checked_str}",
                )
            ).to_have_count(1, timeout=timeout)

        # Could not find the reason why. Raising the original error.
        raise e


def expect_locator_values_in_list(
    *,
    page: Page,
    loc_container: Locator,
    el_type: Locator | str,
    arr_name: str,
    arr: ListPatternOrStr,
    is_checked: bool | MISSING_TYPE = MISSING,
    timeout: Timeout = None,
    key: str = "value",
    alt_verify: bool = False,
) -> None:
    """
    Expect the locator to contain the values in the list.

    The matching values must exist and be in order. No other matching values will be
    allowed within the container.

    Parameters
    ----------
    page
        Playwright `Page` of the Shiny app.
    loc_container
        The container locator.
    el_type
        The element type locator.
    arr_name
        The array name.
    arr
        The expected values.
    is_checked
        Whether the elements are checked. Defaults to `MISSING`.
    timeout
        The timeout for the expectation. Defaults to `None`.
    key
        The key. Defaults to `"value"`.
    alt_verify
        Determines if multiple expectations should be performed.
        Defaults to `False`, a single (and very complicated) locator is asserted.
        `True` will perform multiple assertions, which have the possibility of being invalid.
        Use in playwright bug situations only.
    """
    # Make sure the locator has exactly `arr` values

    # Make sure the locator has len(uniq_arr) input elements
    assert_arr_is_unique(arr, f"`{arr_name}` must be unique")

    if isinstance(el_type, Locator):
        if not isinstance(is_checked, MISSING_TYPE):
            raise RuntimeError(
                "`is_checked` cannot be specified if `el_type` is a Locator"
            )
        loc_item = el_type
    else:
        is_checked_str = checked_css_str(is_checked)
        loc_item = page.locator(f"{el_type}{is_checked_str}")

    # If there are no items, then we should not have any elements
    if len(arr) == 0:
        playwright_expect(loc_container.locator(el_type)).to_have_count(
            0, timeout=timeout
        )
        return
    loc_container_orig = loc_container

    def perform_multiple_assertions():
        # Expecting container to exist (count = 1)
        playwright_expect(loc_container_orig).to_have_count(1, timeout=timeout)

        # Expecting the container to contain {len(arr)} items
        playwright_expect(loc_container_orig.locator(loc_item)).to_have_count(
            len(arr), timeout=timeout
        )

        for item, i in zip(arr, range(len(arr))):
            # Expecting item `{i}` to be `{item}`
            playwright_expect(
                loc_container_orig.locator(loc_item).nth(i)
            ).to_have_attribute(key, item, timeout=timeout)
        return

    if alt_verify:
        # Accordion has issues where the single locator assertion waits forever within playwright.
        # Perform multiple assertions until playwright fixes bug.
        perform_multiple_assertions()
        return

    # Find all items in set
    for item, i in zip(arr, range(len(arr))):
        # Get all elements of type
        has_locator = loc_item
        # Get the `n`th matching element
        has_locator = has_locator.nth(i)
        # Make sure that element has the correct attribute value
        has_locator = has_locator.locator(
            f"xpath=self::*[{_xpath_match_str(key, item)}]"
        )

        # Given the container, make sure it contains this locator
        loc_container = loc_container.locator(
            # Return self
            "xpath=.",
            has=has_locator,
        )

    # Make sure other items are not in set
    # If we know all elements are contained in the container,
    # and all elements all unique, then it should have a count of `len(arr)`
    loc_inputs = loc_container.locator(loc_item)
    try:
        playwright_expect(loc_inputs).to_have_count(len(arr), timeout=timeout)
    except AssertionError as e:
        # Debug expectations
        perform_multiple_assertions()

        # Could not find the reason why. Raising the original error.
        raise e



from __future__ import annotations

import base64
import io
import warnings
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union, cast

from ..types import ImgData, PlotnineFigure
from ._coordmap import get_coordmap, get_coordmap_plotnine

TryPlotResult = Tuple[bool, Union[ImgData, None]]


if TYPE_CHECKING:
    from matplotlib.figure import Figure


class PlotSizeInfo:
    """This class carries information from the render.plot transformer to the logic that
    actually renders the plot to PNG. It also encapsulates the tricky logic for figuring
    out what the image size and attributes should be."""

    def __init__(
        self,
        container_size_px_fn: tuple[Callable[[], float], Callable[[], float]],
        user_specified_size_px: tuple[float | None, float | None],
        pixelratio: float,
    ):
        """
        Parameters
        ----------
        container_size_px_fn
            A tuple of two functions that return the width and height of the container,
            in pixels. If the user specified an explicit width/height on the
            @render.plot decorator, then these functions should not be called, as doing
            so will take a reactive dependency on that dimension.
        user_specified_size_px
            A tuple of two floats, or None. If the user specified an explicit
            width/height on the @render.plot decorator, then the corresponding float
            will be that value. Otherwise, it will be None.
        pixelratio
            The device pixel ratio that was detected from the client.
        """
        self._container_size_px_fn = container_size_px_fn
        self.user_specified_size_px = user_specified_size_px
        self.pixelratio = pixelratio

    def get_img_size_px(
        self,
        fig_initial_size_inches: tuple[float, float],
        fig_result_size_inches: tuple[float, float],
        dpi: float,
    ) -> tuple[float, float, str, str]:
        """
        Determines the desired size of the image in logical pixels (not doubling
        resolution for retina displays--the caller needs to worry about that), and the
        width and height attributes that should be set on the <img> tag.

        Parameters
        ----------
        fig_initial_size_inches
            The default matplotlib/plotnine figure size.
        fig_result_size_inches
            The size of the matplotlib/plotnine figure that the user created. If this is
            different than fig_initial_size_inches, we'll respect it. Otherwise, we'll
            use the container size instead.
        dpi
            The desired DPI of the image.

        Returns
        -------
        :
            A tuple of (width, height, width_attr, height_attr), where width and height
            are the desired size of the image in CSS pixels, and width_attr and
            height_attr are the width and height attributes that should be set on the
            <img> tag. The width and height attributes may differ from the width and
            height values because when we use the container size as the image size, we
            set the width/height attributes to 100% so that the image will continuously
            fill the container even as the container is resized (i.e. the re-render
            hasn't happened yet).
        """
        w, w_attr = self._get_img_size_px(
            0,
            fig_initial_size_inches[0],
            fig_result_size_inches[0],
            dpi,
        )
        h, h_attr = self._get_img_size_px(
            1,
            fig_initial_size_inches[1],
            fig_result_size_inches[1],
            dpi,
        )
        return (w, h, w_attr, h_attr)

    def _get_img_size_px(
        self,
        i: int,
        fig_initial_size_inches: float,
        fig_result_size_inches: float,
        dpi: float,
    ) -> tuple[float, str]:
        # If user specified explicit width/height on @render.plot decorator, set the img
        # size to that exactly. We assume there's some reason they wanted that exact size.
        user_specified_size_px = self.user_specified_size_px[i]
        if user_specified_size_px is not None:
            if user_specified_size_px == 0:
                # If the explicit size is 0, we'll respect the user's figure size.
                native_size = fig_result_size_inches * dpi
                return native_size, f"{native_size}px"
            else:
                return user_specified_size_px, f"{user_specified_size_px}px"

        # If the user didn't specify an explicit size on @render.plot then assume that
        # they're filling the container, in which case we set the img size to 100% in
        # order to have nicer resize behavior.
        #
        # Retrieve the container size, taking a reactive dependency
        container_size_px = self._container_size_px_fn[i]()
        return container_size_px, "100%"


# Try to render a matplotlib object (or the global figure, if it's been used). If `fig`
# is not a matplotlib object, return (False, None). If there's an error in rendering,
# return None. If successful in rendering, return an ImgData object.
def try_render_matplotlib(
    x: object,
    *,
    plot_size_info: PlotSizeInfo,
    allow_global: bool,
    alt: Optional[str],
    **kwargs: object,
) -> TryPlotResult:
    fig = get_matplotlib_figure(x, allow_global)

    if fig is None:
        return (False, None)

    try:
        import matplotlib
        import matplotlib.pyplot as plt  # pyright: ignore[reportUnusedImport] # noqa: F401

        pixelratio = plot_size_info.pixelratio

        fig_initial_size_inches = cast_to_size_tuple(plt.rcParams["figure.figsize"])

        fig_result_size_inches = cast_to_size_tuple(
            fig.get_size_inches(),  # pyright: ignore[reportUnknownMemberType]
        )

        ppi_out = get_desired_dpi_from_fig(fig)

        width, height, width_attr, height_attr = plot_size_info.get_img_size_px(
            fig_initial_size_inches, fig_result_size_inches, ppi_out
        )
        fig.set_size_inches(
            width / ppi_out,
            height / ppi_out,
        )
        fig.set_dpi(ppi_out * pixelratio)

        # Calculating accurate coordinate mappings requires that the layout engine
        # (if there is one) adjusts the figure's subplot parameters.
        # e.g. "tight" layout.
        # When there is no layout engine, "tight" layout is often helpful
        layout_engine = None
        # get_layout_engine was added in matplotlib 3.6
        if hasattr(fig, "get_layout_engine"):
            layout_engine = fig.get_layout_engine()
            if layout_engine:
                if not layout_engine.adjust_compatible:
                    # In most cases, this branch will override the constained layout.
                    # which is usually a very deliberate choice by the user
                    fig.set_layout_engine(  # pyright: ignore[reportUnknownMemberType]
                        layout="tight"
                    )
                    warnings.warn(
                        f"'{type(layout_engine)}' layout engine is not compatible with shiny. "
                        "The figure layout has been changed to tight.",
                        stacklevel=1,
                    )
            else:
                fig.set_layout_engine(  # pyright: ignore[reportUnknownMemberType]
                    layout="tight"
                )
        else:
            # This branch needed for matplotlib <3.6. Eventually we will be able to
            # remove this code path.

            # Suppress the message `UserWarning: The figure layout has changed to tight`
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    category=UserWarning,
                    message="The figure layout has changed to tight",
                )
            plt.tight_layout()  # pyright: ignore[reportUnknownMemberType]

        with io.BytesIO() as buf:
            fig.savefig(  # pyright: ignore[reportUnknownMemberType]
                buf,
                format="png",
                dpi=ppi_out * pixelratio,
                **kwargs,  # pyright: ignore[reportArgumentType, reportGeneralTypeIssues]
            )
            buf.seek(0)
            data = base64.b64encode(buf.read())
            data_str = data.decode("utf-8")

        # Calculating accurate coordinate mappings requires the figure to be
        # drawn/saved first, which runs the layout engine.
        coordmap = get_coordmap(fig)

        res: ImgData = {
            "src": "data:image/png;base64," + data_str,
            "width": width_attr,
            "height": height_attr,
        }

        if alt is not None:
            res["alt"] = alt

        if coordmap is not None:
            res["coordmap"] = coordmap

        return (True, res)

    finally:
        import matplotlib.pyplot

        matplotlib.pyplot.close(fig)  # pyright: ignore[reportUnknownMemberType]


def get_matplotlib_figure(
    x: object, allow_global: bool
) -> Figure | None:  # pyright: ignore
    import matplotlib.pyplot as plt
    from matplotlib.animation import Animation
    from matplotlib.artist import Artist
    from matplotlib.figure import Figure

    # Detect usage of pyplot global figure
    # TODO: Might be good to detect non-empty plt.get_fignums() before we call the user
    #   function, which would mean we will false-positive here. Maybe we warn in that
    #   case, maybe we ignore gcf(), maybe both.
    if x is None and len(plt.get_fignums()) > 0:
        if allow_global:
            return plt.gcf()
        else:
            # Must close the global figure so we don't stay in this state forever
            plt.close(plt.gcf())  # pyright: ignore[reportUnknownMemberType]
            raise RuntimeError(
                "matplotlib.pyplot cannot be used from an async render function; "
                "please use matplotlib's object-oriented interface instead"
            )

    if isinstance(x, Figure):
        return x

    if isinstance(x, Animation):
        raise RuntimeError(
            "Matplotlib's Animation class isn't supported by @render.plot. "
            + "Consider explictly saving the animation to a file and "
            + "then using @render.image instead to render it."
        )

    # Libraries like pandas, xarray, etc have plot() methods that can return a wide
    # array pf mpl classes, like Lines2D, Subplots, Axes, etc. The Artist ABC class
    # should cover most, if not all, of these (it doesn't cover Animation, though).
    # https://matplotlib.org/stable/api/artist_api.html
    if isinstance(x, Artist):
        return x.get_figure()

    # Some other custom figure-like classes such as seaborn.axisgrid.FacetGrid attach
    # their figure as an attribute
    fig = getattr(x, "figure", None)
    if isinstance(fig, Figure):
        return fig

    # Sometimes generic plot() methods will return an iterable of Artists,
    # If they all refer to the same figure, then it seems reasonable to use it
    # https://docs.xarray.dev/en/latest/user-guide/plotting.html#dimension-along-y-axis
    if isinstance(x, (list, tuple)):
        figs = [get_matplotlib_figure(y, allow_global) for y in cast(List[Any], x)]
        if len(set(figs)) == 1:
            return figs[0]

    return None


def try_render_pil(
    x: object,
    *,
    plot_size_info: PlotSizeInfo,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    import PIL.Image

    if not isinstance(x, PIL.Image.Image):
        return (False, None)

    with io.BytesIO() as buf:
        x.save(  # pyright: ignore[reportUnknownMemberType]
            buf,
            format="PNG",
            **kwargs,  # pyright: ignore[reportArgumentType,reportGeneralTypeIssues]
        )
        buf.seek(0)
        data = base64.b64encode(buf.read())
        data_str = data.decode("utf-8")

    width_attr = plot_size_info.user_specified_size_px[0]
    width_attr = f"{width_attr}px" if width_attr is not None else "100%"
    height_attr = plot_size_info.user_specified_size_px[1]
    height_attr = f"{height_attr}px" if height_attr is not None else "100%"

    res: ImgData = {
        "src": "data:image/png;base64," + data_str,
        "width": width_attr,
        "height": height_attr,
        "style": "object-fit:contain",
    }

    if alt is not None:
        res["alt"] = alt

    return (True, res)


def try_render_plotnine(
    x: object,
    *,
    plot_size_info: PlotSizeInfo,
    alt: Optional[str] = None,
    **kwargs: object,
) -> TryPlotResult:
    import plotnine.options as p9options
    from plotnine.ggplot import ggplot

    if not isinstance(x, ggplot):
        return (False, None)

    fig_initial_size_inches = p9options.figure_size

    x = cast(PlotnineFigure, x)

    fig_result_size_inches = fig_initial_size_inches
    figure_size = x.theme.themeables.get("figure_size")
    if figure_size is not None:
        result_size = figure_size.properties.get("value")
        if result_size is not None:
            fig_result_size_inches = result_size
    ppi = p9options.dpi
    figure_dpi = x.theme.themeables.get("dpi")
    if figure_dpi is not None:
        result_dpi = figure_dpi.properties.get("value")
        if result_dpi is not None:
            ppi = result_dpi

    w, h, w_attr, h_attr = plot_size_info.get_img_size_px(
        fig_initial_size_inches, fig_result_size_inches, ppi
    )

    with io.BytesIO() as buf:
        if not hasattr(x, "save_helper"):
            raise RuntimeError(
                "plotnine>=0.10.1 is required to render plotnine plots in Shiny"
            )
        res = x.save_helper(  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportUnknownVariableType, reportGeneralTypeIssues]
            filename=buf,
            format="png",
            units="in",
            dpi=ppi * plot_size_info.pixelratio,
            width=w / ppi,
            height=h / ppi,
            verbose=False,
            **kwargs,
        )
        res.figure.savefig(  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportGeneralTypeIssues]
            **res.kwargs  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue, reportGeneralTypeIssues]
        )
        buf.seek(0)
        data = base64.b64encode(buf.read())
        data_str = data.decode("utf-8")

    # Calculating accurate coordinate mappings requires the figure to be
    # drawn/saved first, which runs the layout engine.
    coordmap = get_coordmap_plotnine(
        x,
        res.figure,  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType, reportAttributeAccessIssue, reportGeneralTypeIssues]
    )

    res: ImgData = {
        "src": "data:image/png;base64," + data_str,
        "width": w_attr,
        "height": h_attr,
    }

    if alt is not None:
        res["alt"] = alt

    if coordmap is not None:
        res["coordmap"] = coordmap

    return (True, res)


# This is a weird one... the default dpi is not set to rcParam["figure.dpi"], but rather
# to rcParam["figure.dpi"] * fig.canvas.device_pixel_ratio (which is 2.0 on my Mac with
# the 'MacOSX' mpl backend). We want to undo that scaling, as it makes the text
# ridiculously large.
#
# One negative consequence of this logic: if the user intentionally set the dpi to
# rcParam * device_pixel_ratio, we're going to ignore it.
def get_desired_dpi_from_fig(fig: Figure) -> float:
    ppi_out = fig.get_dpi()

    if fig.canvas.device_pixel_ratio != 1 and hasattr(fig, "_original_dpi"):
        if fig._original_dpi == ppi_out / fig.canvas.device_pixel_ratio:  # type: ignore
            return cast(float, fig._original_dpi)  # type: ignore
    return ppi_out


def cast_to_size_tuple(lst: Any) -> tuple[float, float]:
    return cast(Tuple[float, float], tuple(lst))



# pyright: basic

from custom_component import (
    input_custom_component,
    output_custom_component,
    render_custom_component,
)

from shiny import App, ui

app_ui = ui.page_fluid(
    ui.h2("Color picker"),
    input_custom_component("color"),
    ui.br(),
    ui.h2("Output color"),
    output_custom_component("value"),
)


def server(input, output, session):
    @render_custom_component
    def value():
        print("Calculating value")
        return input.color()


app = App(app_ui, server, debug=True)



from __future__ import annotations

import pandas as pd

from shiny import App, Inputs, Outputs, Session, render, ui

df = pd.DataFrame(
    dict(
        id=["one", "two", "three"],
        fname=["Alice", "Bob", "Charlie"],
        lname=["Smith", "Jones", "Brown"],
    )
).set_index(  # type: ignore
    "id",
    drop=False,
)

app_ui = ui.page_fluid(
    ui.p("Select rows in the grid, make sure the selected rows appear below."),
    ui.output_data_frame("grid"),
    ui.p(
        "Selected data: ",
        ui.output_data_frame("detail"),
    ),
    ui.p(
        "Selected rows: ",
        ui.output_code("selected_rows"),
    ),
    class_="p-3",
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.data_frame
    def grid():
        return render.DataGrid(
            df,
            selection_mode="rows",
            height=350,
        )

    @render.data_frame
    def detail():
        return grid.data_view(selected=True)

    @render.code
    def selected_rows():
        return str(grid.cell_selection()["rows"])


app = App(app_ui, server, debug=False)



from __future__ import annotations

__all__ = ("input_file",)

from typing import Literal, Optional

from htmltools import Tag, TagChild, css, div, span, tags

from .._docstring import add_example
from .._namespaces import resolve_id
from ._utils import shiny_input_label


@add_example()
def input_file(
    id: str,
    label: TagChild,
    *,
    multiple: bool = False,
    accept: Optional[str | list[str]] = None,
    width: Optional[str] = None,
    button_label: str = "Browse...",
    placeholder: str = "No file selected",
    capture: Optional[Literal["environment", "user"]] = None,
) -> Tag:
    """
    Create a file upload control that can be used to upload one or more files.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    multiple
        Whether the user should be allowed to select and upload multiple files at once.
    accept
        Unique file type specifier(s) which give the browser a hint as to the type of
        file the server expects. Many browsers use this to prevent the user from
        selecting an invalid file. Examples of valid values include a case insensitive
        extension (e.g. ``.csv`` or ``.rds``), a valid MIME type (e.g. ``text/plain`` or
        ``application/pdf``) or one of ``audio/*``, ``video/*``, or ``image/*`` meaning
        any audio, video, or image type, respectively.
    width
        The CSS width, e.g. '400px', or '100%'
    button_label
        The label used on the button.
    placeholder
        The text to show on the input before a file has been uploaded.
    capture
        On mobile devices, this can be used to open the device's camera for input. If
        "environment", it will open the rear-facing camera. If "user", it will open the
        front-facing camera. By default, it will accept either still photos or video. To
        accept only still photos, use ``accept="image/*"``; to accept only video, use
        ``accept="video/*"``.

    Returns
    -------
    :
        A UI element.

    Notes
    -----

    ::: {.callout-note title="Server value"}
    A list of dictionaries (one for each file upload) with the following keys:

    * ``name``: The filename provided by the web browser. This is *not* the path to read
        to get at the actual data that was uploaded (see 'datapath').
    * ``size``: The size of the uploaded data, in bytes.
    * ``type``: The MIME type reported by the browser (for example, 'text/plain'), or
        empty string if the browser didn't know.
    * ``datapath``: The path to a temp file that contains the data that was uploaded.
        This file may be deleted if the user performs another upload operation.
    :::

    See Also
    --------
    * :func:`~shiny.ui.download_button`
    """

    if isinstance(accept, str):
        accept = [accept]

    resolved_id = resolve_id(id)
    btn_file = span(
        button_label,
        tags.input(
            id=resolved_id,
            name=resolved_id,
            type="file",
            multiple="multiple" if multiple else None,
            accept=",".join(accept) if accept else None,
            capture=capture,
            # Don't use "display: none;" style, which causes keyboard accessibility issue; instead use the following workaround: https://css-tricks.com/places-its-tempting-to-use-display-none-but-dont/
            style="position: absolute !important; top: -99999px !important; left: -99999px !important;",
            class_="shiny-input-file",
        ),
        class_="btn btn-default btn-file",
    )
    return div(
        shiny_input_label(resolved_id, label),
        div(
            tags.label(btn_file, class_="input-group-btn input-group-prepend"),
            tags.input(
                type="text",
                class_="form-control",
                placeholder=placeholder,
                readonly="readonly",
            ),
            class_="input-group",
        ),
        div(
            div(class_="progress-bar"),
            id=resolved_id + "_progress",
            class_="progress active shiny-file-input-progress",
        ),
        class_="form-group shiny-input-container",
        style=css(width=width),
    )



from __future__ import annotations

import base64
import datetime
import os
from pathlib import Path

import pytest
from examples.example_apps import reruns, reruns_delay
from mod_state import expect_default_mod_state, expect_mod_state
from playwright.sync_api import Page

from shiny._utils import guess_mime_type
from shiny.playwright import controller
from shiny.run import ShinyAppProc

img_path = Path(__file__).parent / "imgs"
penguin_imgs = [str(img_path / img) for img in os.listdir(img_path)]


def expect_outputs(page: Page, module_id: str, letter: str, count: int):
    def resolve_id(id: str):
        if module_id:
            return f"{module_id}-{id}"
        return id

    dataframe = controller.OutputDataFrame(page, resolve_id("out_data_frame"))
    dataframe.expect_nrow(count + 1)

    controller.OutputText(page, resolve_id("out_text")).expect_value(
        f"Output text content. `input.radio_buttons()`: `{letter}`"
    )
    controller.OutputTextVerbatim(page, resolve_id("out_text_verbatim")).expect_value(
        f"Output text verbatim content. `input.radio_buttons()`: `{letter}`"
    )
    controller.OutputTable(page, resolve_id("out_table")).expect_nrow(count + 1)

    # Mirrors ImageTransformer implementation
    src = penguin_imgs[count]
    with open(src, "rb") as f:
        data = base64.b64encode(f.read())
        data_str = data.decode("utf-8")
    content_type = guess_mime_type(src)
    img_src = f"data:{content_type};base64,{data_str}"
    # Done Mirror
    controller.OutputImage(page, resolve_id("out_image")).expect_img_src(img_src)

    # TODO-future; Test OutputPlot
    # controller.OutputPlot(page, resolve_id("out_plot")).

    controller.OutputUi(page, resolve_id("out_ui")).expect.to_have_text(
        f"Output UI content. input.radio_buttons(): {letter}"
    )


def expect_labels(page: Page, module_id: str):
    def resolve_id(id: str):
        if module_id:
            return f"{module_id}-{id}"
        return id

    controller.InputNumeric(page, resolve_id("input_numeric")).expect_label("Numeric")
    controller.InputText(page, resolve_id("input_text")).expect_label("Text")
    controller.InputTextArea(page, resolve_id("input_text_area")).expect_label(
        "Text area"
    )
    controller.InputSelect(page, resolve_id("input_select")).expect_label("Select")
    controller.InputSelectize(page, resolve_id("input_selectize")).expect_label(
        "Selectize"
    )
    controller.InputCheckbox(page, resolve_id("input_checkbox")).expect_label(
        "Checkbox"
    )
    controller.InputSwitch(page, resolve_id("input_switch")).expect_label("Switch")
    controller.InputCheckboxGroup(
        page, resolve_id("input_checkbox_group")
    ).expect_label("Checkbox group")
    controller.InputRadioButtons(page, resolve_id("input_radio_buttons")).expect_label(
        "Radio buttons"
    )
    controller.InputFile(page, resolve_id("input_file")).expect_label("File")
    controller.InputSlider(page, resolve_id("input_slider")).expect_label("Slider")
    controller.InputDate(page, resolve_id("input_date")).expect_label("Date")
    controller.InputDateRange(page, resolve_id("input_date_range")).expect_label(
        "Date range"
    )


def expect_default_outputs(page: Page, module_id: str):
    expect_outputs(page, module_id, "a", 0)


# Sidebars do not seem to work on webkit. Skipping test on webkit
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_module_support(page: Page, local_app: ShinyAppProc) -> None:
    page.set_viewport_size({"width": 3000, "height": 6000})
    page.goto(local_app.url)

    # Verify reset state
    for mod_id in ("", "mod1", "mod2"):
        expect_default_mod_state(page, mod_id)
        expect_default_outputs(page, mod_id)
        expect_labels(page, mod_id)

    # Click x3 `update_mod2`
    update_mod2 = controller.InputActionButton(page, "update_mod2")
    for i in range(3):
        update_mod2.click()
        controller.InputActionButton(page, "mod2-input_action_button").click()
        controller.InputActionLink(page, "mod2-input_action_link").click()
        with page.expect_download() as download_button_info:
            download_button = controller.DownloadButton(page, "mod2-download_button")
            download_button.click()
            download = download_button_info.value
            # wait for download to complete
            download_path = download.path()
            assert download.suggested_filename == "download_button-mod2.csv"
            assert download_path is not None
            with open(download_path, "r") as f:
                assert f.read() == f"session,type,count\nmod2,button,{i + 1}\n"
        with page.expect_download() as download_link_info:
            download_link = controller.DownloadLink(page, "mod2-download_link")
            download_link.click()
            download = download_link_info.value
            # wait for download to complete
            download_path = download.path()
            assert download.suggested_filename == "download_link-mod2.csv"
            assert download_path is not None
            with open(download_path, "r") as f:
                assert f.read() == f"session,type,count\nmod2,link,{i + 1}\n"

    controller.InputFile(page, "mod2-input_file").set(
        Path(__file__).parent / "test_file.txt"
    )

    # Make sure the global session and first module was not affected
    expect_default_mod_state(page, "")
    expect_default_outputs(page, "")
    expect_default_mod_state(page, "mod1")
    expect_default_outputs(page, "mod1")

    # Make sure the second module was affected
    expect_mod_state(
        page,
        module_id="mod2",
        sidebar=False,
        accordion=("d",),
        popover=True,
        tooltip=True,
        input_action_button=3,
        input_action_link=3,
        input_file="test_file.txt",
        input_checkbox=True,
        input_checkbox_group=("a", "b", "c"),
        input_date=datetime.date(2023, 8, 27),
        input_date_range=(datetime.date(2023, 8, 28), datetime.date(2023, 8, 30)),
        input_numeric=3,
        input_password="password3",
        input_radio_buttons="d",
        input_select="d",
        input_selectize="d",
        input_slider=3,
        input_switch=True,
        input_text="text3",
        input_text_area="text_area3",
        navset_bar="d",
        navset_card_pill="d",
        navset_card_tab="d",
        navset_hidden="d",
        navset_pill="d",
        navset_tab="d",
    )
    expect_outputs(page, "mod2", "d", 3)



# ------------------------------------------------------------------------------------
# A basic Shiny Chat example powered by Anthropic's Claude model.
# To run it, you'll need an Anthropic API key.
# To get one, follow the instructions at https://docs.anthropic.com/en/api/getting-started
# ------------------------------------------------------------------------------------
import os

from anthropic import AsyncAnthropic
from app_utils import load_dotenv

from shiny.express import ui

# Either explicitly set the ANTHROPIC_API_KEY environment variable before launching the
# app, or set them in a file named `.env`. The `python-dotenv` package will load `.env`
# as environment variables which can later be read by `os.getenv()`.
load_dotenv()
llm = AsyncAnthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))

# Set some Shiny page options
ui.page_opts(
    title="Hello Anthropic Claude Chat",
    fillable=True,
    fillable_mobile=True,
)

# Create and display empty chat
chat = ui.Chat(id="chat")
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def _():
    # Get messages currently in the chat
    messages = chat.messages(format="anthropic")
    # Create a response message stream
    response = await llm.messages.create(
        model="claude-3-5-sonnet-latest",
        messages=messages,
        stream=True,
        max_tokens=1000,
    )
    # Append the response stream into the chat
    await chat.append_message_stream(response)



import pandas as pd

from shiny import App, Inputs, Outputs, Session, reactive, render, ui
from shiny.types import FileInfo

app_ui = ui.page_fluid(
    ui.input_file("file1", "Choose CSV File", accept=[".csv"], multiple=False),
    ui.input_checkbox_group(
        "stats",
        "Summary Stats",
        choices=["Row Count", "Column Count", "Column Names"],
        selected=["Row Count", "Column Count", "Column Names"],
    ),
    ui.output_table("summary"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @reactive.calc
    def parsed_file():
        file: list[FileInfo] | None = input.file1()
        if file is None:
            return pd.DataFrame()
        return pd.read_csv(  # pyright: ignore[reportUnknownMemberType]
            file[0]["datapath"]
        )

    @render.table
    def summary():
        df = parsed_file()

        if df.empty:
            return pd.DataFrame()

        # Get the row count, column count, and column names of the DataFrame
        row_count = df.shape[0]
        column_count = df.shape[1]
        names = df.columns.tolist()
        column_names = ", ".join(str(name) for name in names)

        # Create a new DataFrame to display the information
        info_df = pd.DataFrame(
            {
                "Row Count": [row_count],
                "Column Count": [column_count],
                "Column Names": [column_names],
            }
        )

        # input.stats() is a list of strings; subset the columns based on the selected
        # checkboxes
        return info_df.loc[:, input.stats()]


app = App(app_ui, server)



# Import the custom renderer implementations
from renderers import render_capitalize, render_upper

from shiny.express import input, ui

ui.h1("Capitalization renderer")
ui.input_text("caption", "Caption:", "Data summary")

"@render_upper:"


# Hovering over `@render_upper` will display the class documentation
@render_upper
def upper():
    return input.caption()


"@render_upper():"


# Hovering over `@render_upper` will display the class documentation as there is no
# `__init__()` documentation
@render_upper()
def upper_with_paren():
    return input.caption()


"@render_capitalize:"


# Hovering over `@render_capitalize` will display the class documentation
@render_capitalize
def cap_upper():
    return input.caption()


"@render_capitalize(to='lower'): "


# Hovering over `@render_capitalize` will display the `__init__()` documentation
@render_capitalize(to_case="lower")
def cap_lower():
    return input.caption()



from __future__ import annotations

import json
from typing import Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, tags

from .._docstring import add_example
from .._namespaces import resolve_id_or_none
from ._tag import consolidate_attrs
from ._web_component import web_component


@add_example()
def tooltip(
    trigger: TagChild,
    *args: TagChild | TagAttrs,
    id: Optional[str] = None,
    placement: Literal["auto", "top", "right", "bottom", "left"] = "auto",
    options: Optional[dict[str, object]] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Add a tooltip to a UI element.

    Display additional information when focusing (or hovering over) a UI element.

    Parameters
    ----------
    trigger
        A UI element (i.e., :class:`~htmltools.Tag`) to serve as the tooltips trigger.
        It's good practice for this element to be a keyboard-focusable and interactive
        element (e.g., :func:`~shiny.ui.input_action_button`,
        :func:`~shiny.ui.input_action_link`, etc.) so that the tooltip is accessible to
        keyboard and assistive technology users.
    *args
        Contents to the tooltip's body. Or tag attributes that are supplied to the
        resolved :class:`~htmltools.Tag` object.
    id
        A character string. Required to reactively respond to the visibility of the
        tooltip (via the `input[id]` value) and/or update the visibility/contents of the
        tooltip.
    placement
        The placement of the tooltip relative to its trigger.
    options
        A list of additional [Bootstrap
        options](https://getbootstrap.com/docs/5.3/components/tooltips/#options).

    Details
    -------

    If `trigger` yields multiple HTML elements (e.g., a :class:`~htmltools.TagList` or
    complex [`shinywidgets`](https://github.com/posit-dev/py-shinywidgets) object), the
    last HTML element is used as the trigger. If the `trigger` should contain all of
    those elements, wrap the object in a :func:`~shiny.ui.tags.div` or
    :func:`~shiny.ui.tags.span`.

    Accessibility of Tooltip Triggers
    ---------------------------------

    Because the user needs to interact with the `trigger` element to see the `tooltip`,
    it's best practice to use an element that is typically accessible via keyboard
    interactions, like a button or a link.

    If you use a non-interactive element, like a `<span>` or text, `tooltip()` will
    automatically add the `tabindex="0"` attribute to the trigger element to make sure
    that users can reach the element with the keyboard. This means that in most cases
    you can use any element you want as the trigger.

    One place where it's important to consider the accessibility of the trigger is when
    using an icon without any accompanying text. In these cases, many icon elements are
    created with the assumption that the icon is decorative, which will make it
    inaccessible to users of assistive technologies.

    When using an icon as the primary trigger, ensure that the icon does not have
    `aria-hidden="true"` or `role="presentation"` attributes. Icon packages typically
    provide a way to specify a title for the icon, as well as a way to specify that the
    icon is not decorative. The title should be a short description of the purpose of
    the trigger, rather than a description of the icon itself.

    For example:

    ```python
    icon_title = "About tooltips"
    def bs_info_icon(title: str):
        # Enhanced from https://rstudio.github.io/bsicons/ via `bsicons::bs_icon("info-circle", title = icon_title)`
        return ui.HTML(f'<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16" class="bi bi-info-circle " style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img" ><title>{title}</title><path d="M8 15A7 7 0 1 1 8 1a7 7 0 0 1 0 14zm0 1A8 8 0 1 0 8 0a8 8 0 0 0 0 16z"></path><path d="m8.93 6.588-2.29.287-.082.38.45.083c.294.07.352.176.288.469l-.738 3.468c-.194.897.105 1.319.808 1.319.545 0 1.178-.252 1.465-.598l.088-.416c-.2.176-.492.246-.686.246-.275 0-.375-.193-.304-.533L8.93 6.588zM9 4.5a1 1 0 1 1-2 0 1 1 0 0 1 2 0z"></path></svg>')

    ui.tooltip(
        bs_info_icon(icon_title),
        "Text shown in the tooltip."
    )
    ```

    ```python
    icon_title = "About tooltips"
    def fa_info_circle(title: str):
        # Enhanced from https://rstudio.github.io/fontawesome/ via `fontawesome::fa("info-circle", a11y = "sem", title = icon_title)`
        return ui.HTML(f'<svg aria-hidden="true" role="img" viewBox="0 0 512 512" style="height:1em;width:1em;vertical-align:-0.125em;margin-left:auto;margin-right:auto;font-size:inherit;fill:currentColor;overflow:visible;position:relative;"><title>{title}</title><path d="M256 512A256 256 0 1 0 256 0a256 256 0 1 0 0 512zM216 336h24V272H216c-13.3 0-24-10.7-24-24s10.7-24 24-24h48c13.3 0 24 10.7 24 24v88h8c13.3 0 24 10.7 24 24s-10.7 24-24 24H216c-13.3 0-24-10.7-24-24s10.7-24 24-24zm40-208a32 32 0 1 1 0 64 32 32 0 1 1 0-64z"/></svg>')
    ui.tooltip(
        fa_info_circle(icon_title),
        "Text shown in the tooltip."
    )
    ```

    See Also
    --------

    * [Bootstrap tooltips documentation](https://getbootstrap.com/docs/5.3/components/tooltips/)
    """

    # * If you're using [bsicons::bs_icon()], provide a `title`.
    # * If you're using [fontawesome::fa()], set `a11y = "sem"` and provide a `title`.

    # Theming/Styling
    # ---------------
    #
    # Like other bslib components, tooltips can be themed by supplying [relevant theming
    # variables](https://rstudio.github.io/bslib/articles/bs5-variables.html#tooltip-bg)
    # to [bs_theme()], which effects styling of every tooltip on the page. To style a
    # _specific_ tooltip differently from other tooltips, utilize the `customClass`
    # option:
    #
    # ```
    # tooltip(
    #     "Trigger", "Tooltip message",
    #     options = {"customClass": "my-pop"}
    # )
    # ```
    #
    # And then add relevant rules to [bs_theme()] via [bs_add_rules()]:
    #
    # ```
    # bs_theme() |> bs_add_rules(".my-pop { max-width: none; }")
    # ```

    attrs, children = consolidate_attrs(*args, **kwargs)

    if len(children) == 0:
        raise RuntimeError("At least one value must be provided to `*args: TagChild`")

    res = web_component(
        "bslib-tooltip",
        {
            "id": resolve_id_or_none(id),
            "placement": placement,
            "options": json.dumps(options) if options else None,
        },
        attrs,
        # Use display:none instead of <template> since shiny.js
        # doesn't bind to the contents of the latter
        tags.template(*children, {"style": "display:none;"}),
        trigger,
    )

    return res



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from shiny._static import (
    _installed_shinylive_versions,
    get_default_shinylive_dir,
    remove_shinylive_local,
)


@contextmanager
def run_within_dir(path: Path) -> Generator[None, None, None]:
    """
    Utility function to run some code within a directory, and change back to the current directory afterwards.
    Example usage:
    ```
    with run_within_dir(directory):
        some_code()
    ```
    """
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


def test_get_default_shinylive_dir():
    assert isinstance(get_default_shinylive_dir(), Path)


def test_remove_shinylive_local(tmp_path: Path):
    with run_within_dir(tmp_path):
        shinylive_path = Path("shinylive")
        shinylive_path.mkdir()
        assert os.path.exists(shinylive_path)
        remove_shinylive_local(shinylive_path)
        assert not os.path.exists(shinylive_path)


def test_installed_shinylive_versions(tmp_path: Path):
    with run_within_dir(tmp_path):
        Path("shinylive/shinylive-foo").mkdir(parents=True, exist_ok=True)
        Path("shinylive/shinylive-bar").mkdir(parents=True, exist_ok=True)
        Path("shinylive/shinylive-bar/nested_directory").mkdir(
            parents=True, exist_ok=True
        )

        versions = _installed_shinylive_versions(shinylive_dir=Path("shinylive"))
        assert set(versions) == {"foo", "bar"}



from shiny import reactive, req
from shiny.express import input, ui

with ui.sidebar():
    ui.input_action_button("update", "Update other buttons and link")

with ui.layout_column_wrap():
    ui.input_action_button("goButton", "Go")
    ui.input_action_button("goButton2", "Go 2", icon="🤩")
    ui.input_action_button("goButton3", "Go 3")
    ui.input_action_link("goLink", "Go Link")


@reactive.effect
def _():
    req(input.update())
    # Updates goButton's label and icon
    ui.update_action_button("goButton", label="New label", icon="📅")
    # Leaves goButton2's label unchanged and removes its icon
    ui.update_action_button("goButton2", icon=[])
    # Leaves goButton3's icon, if it exists, unchanged and changes its label
    ui.update_action_button("goButton3", label="New label 3")
    # Updates goLink's label and icon
    ui.update_action_link("goLink", label="New link label", icon="🔗")



from typing import Union

from shiny.express import render, ui

# Set some Shiny page options
ui.page_opts(title="Hello Chat")

# Create a chat instance, with an initial message
chat = ui.Chat(id="chat")

# Display the chat
chat.ui()


# TODO: test with append_message_stream() as well
@chat.transform_assistant_response
def transform(content: str) -> Union[str, ui.HTML]:
    if content == "return HTML":
        return ui.HTML(f"<b>Transformed response</b>: {content}")
    else:
        return f"Transformed response: `{content}`"


@chat.on_user_submit
async def _():
    user = chat.user_input()
    await chat.append_message(user)


"chat.messages():"


@render.code
def message_state():
    return str(chat.messages())


"chat.messages(transform_assistant=True):"


@render.code
def message_state2():
    return str(chat.messages(transform_assistant=True))



from __future__ import annotations

from typing import Literal, TypedDict

from htmltools import HTML

Role = Literal["assistant", "user", "system"]


# TODO: content should probably be [{"type": "text", "content": "..."}, {"type": "image", ...}]
# in order to support multiple content types...
class ChatMessage(TypedDict):
    content: str
    role: Role


# A message once transformed have been applied
class TransformedMessage(TypedDict):
    content_client: str | HTML
    content_server: str
    role: Role
    transform_key: Literal["content_client", "content_server"]
    pre_transform_key: Literal["content_client", "content_server"]


# A message that can be sent to the client
class ClientMessage(ChatMessage):
    content_type: Literal["markdown", "html"]
    chunk_type: Literal["message_start", "message_end"] | None



from shiny import reactive
from shiny.express import input, render, ui

ui.input_password("password", "Password:")
ui.input_action_button("go", "Go")


@render.code
@reactive.event(input.go)
def value():
    return input.password()



from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat_transform_assistant(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "chat")
    message_state = controller.OutputCode(page, "message_state")

    # Wait for app to load
    message_state.expect_value("()", timeout=30 * 1000)

    expect(chat.loc).to_be_visible(timeout=30 * 1000)
    expect(chat.loc_input_button).to_be_disabled()

    chat.set_user_input("foo")
    chat.send_user_input()
    chat.expect_latest_message("Simple response...DONE!", timeout=30 * 1000)

    message_state_expected = tuple(
        [
            {"content": "foo", "role": "user"},
            {"content": "Simple response", "role": "assistant"},
        ]
    )
    message_state.expect_value(str(message_state_expected))



from __future__ import annotations

import inspect
from typing import (
    Any,
    AsyncIterable,
    Awaitable,
    Callable,
    Iterable,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)
from weakref import WeakValueDictionary

from htmltools import HTML, Tag, TagAttrValue, css

from .. import _utils, reactive
from .._deprecated import warn_deprecated
from .._docstring import add_example
from .._namespaces import ResolvedId, resolve_id
from ..session import require_active_session, session_context
from ..types import MISSING, MISSING_TYPE, NotifyException
from ..ui.css import CssUnit, as_css_unit
from ._chat_normalize import normalize_message, normalize_message_chunk
from ._chat_provider_types import (
    AnthropicMessage,
    GoogleMessage,
    LangChainMessage,
    OllamaMessage,
    OpenAIMessage,
    ProviderMessage,
    ProviderMessageFormat,
    as_provider_message,
)
from ._chat_tokenizer import TokenEncoding, TokenizersEncoding, get_default_tokenizer
from ._chat_types import ChatMessage, ClientMessage, TransformedMessage
from ._html_deps_py_shiny import chat_deps
from .fill import as_fill_item, as_fillable_container

__all__ = (
    "Chat",
    "chat_ui",
    "ChatMessage",
)


# TODO: UserInput might need to be a list of dicts if we want to support multiple
# user input content types
TransformUserInput = Callable[[str], Union[str, None]]
TransformUserInputAsync = Callable[[str], Awaitable[Union[str, None]]]
TransformAssistantResponse = Callable[[str], Union[str, HTML, None]]
TransformAssistantResponseAsync = Callable[[str], Awaitable[Union[str, HTML, None]]]
TransformAssistantResponseChunk = Callable[[str, str, bool], Union[str, HTML, None]]
TransformAssistantResponseChunkAsync = Callable[
    [str, str, bool], Awaitable[Union[str, HTML, None]]
]
TransformAssistantResponseFunction = Union[
    TransformAssistantResponse,
    TransformAssistantResponseAsync,
    TransformAssistantResponseChunk,
    TransformAssistantResponseChunkAsync,
]
SubmitFunction = Callable[[], None]
SubmitFunctionAsync = Callable[[], Awaitable[None]]

ChunkOption = Literal["start", "end", True, False]

PendingMessage = Tuple[Any, ChunkOption, Union[str, None]]


@add_example(ex_dir="../api-examples/chat")
class Chat:
    """
    Create a chat interface.

    A UI component for building conversational interfaces. With it, end users can submit
    messages, which will cause a `.on_user_submit()` callback to run. In that callback,
    a response can be generated based on the chat's `.messages()`, and appended to the
    chat using `.append_message()` or `.append_message_stream()`.

    Here's a rough outline for how to implement a `Chat`:

    ```python
    from shiny.express import ui

    # Create and display chat instance
    chat = ui.Chat(id="my_chat")
    chat.ui()

    # Define a callback to run when the user submits a message
    @chat.on_user_submit
    async def _():
        # Get messages currently in the chat
        messages = chat.messages()
        # Create a response message stream
        response = await my_model.generate_response(messages, stream=True)
        # Append the response into the chat
        await chat.append_message_stream(response)
    ```

    In the outline above, `my_model.generate_response()` is a placeholder for
    the function that generates a response based on the chat's messages. This function
    will look different depending on the model you're using, but it will generally
    involve passing the messages to the model and getting a response back. Also, you'll
    typically have a choice to `stream=True` the response generation, and in that case,
    you'll use `.append_message_stream()` instead of `.append_message()` to append the
    response to the chat. Streaming is preferrable when available since it allows for
    more responsive and scalable chat interfaces.

    Parameters
    ----------
    id
        A unique identifier for the chat session. In Shiny Core, make sure this id
        matches a corresponding :func:`~shiny.ui.chat_ui` call in the UI.
    messages
        A sequence of messages to display in the chat. Each message can be a
        dictionary with a `content` and `role` key. The `content` key should contain
        the message text, and the `role` key can be "assistant", "user", or "system".
        Note that system messages are not actually displayed in the chat, but will
        still be stored in the chat's `.messages()`.
    on_error
        How to handle errors that occur in response to user input. When `"unhandled"`,
        the app will stop running when an error occurs. Otherwise, a notification
        is displayed to the user and the app continues to run.

        * `"auto"`: Sanitize the error message if the app is set to sanitize errors,
          otherwise display the actual error message.
        * `"actual"`: Display the actual error message to the user.
        * `"sanitize"`: Sanitize the error message before displaying it to the user.
        * `"unhandled"`: Do not display any error message to the user.
    tokenizer
        The tokenizer to use for calculating token counts, which is required to impose
        `token_limits` in `.messages()`. If not provided, a default generic tokenizer
        is attempted to be loaded from the tokenizers library. A specific tokenizer
        may also be provided by following the `TokenEncoding` (tiktoken or tozenizers)
        protocol (e.g., `tiktoken.encoding_for_model("gpt-4o")`).
    """

    def __init__(
        self,
        id: str,
        *,
        messages: Sequence[Any] = (),
        on_error: Literal["auto", "actual", "sanitize", "unhandled"] = "auto",
        tokenizer: TokenEncoding | None = None,
    ):
        if not isinstance(id, str):
            raise TypeError("`id` must be a string.")

        self.id = resolve_id(id)
        self.user_input_id = ResolvedId(f"{self.id}_user_input")
        self._transform_user: TransformUserInputAsync | None = None
        self._transform_assistant: TransformAssistantResponseChunkAsync | None = None
        self._tokenizer = tokenizer

        # TODO: remove the `None` when this PR lands:
        # https://github.com/posit-dev/py-shiny/pull/793/files
        self._session = require_active_session(None)

        # Default to sanitizing until we know the app isn't sanitizing errors
        if on_error == "auto":
            on_error = "sanitize"
            app = self._session.app
            if app is not None and not app.sanitize_errors:  # type: ignore
                on_error = "actual"

        self.on_error = on_error

        # Chunked messages get accumulated (using this property) before changing state
        self._current_stream_message = ""
        self._current_stream_id: str | None = None
        self._pending_messages: list[PendingMessage] = []

        # If a user input message is transformed into a response, we need to cancel
        # the next user input submit handling
        self._suspend_input_handler: bool = False

        # Keep track of effects so we can destroy them when the chat is destroyed
        self._effects: list[reactive.Effect_] = []

        # Initialize chat state and user input effect
        with session_context(self._session):
            # Initialize message state
            self._messages: reactive.Value[tuple[TransformedMessage, ...]] = (
                reactive.Value(())
            )

            self._latest_user_input: reactive.Value[TransformedMessage | None] = (
                reactive.Value(None)
            )

            # TODO: deprecate messages once we start promoting managing LLM message
            # state through other means
            @reactive.effect
            async def _init_chat():
                for msg in messages:
                    await self.append_message(msg)

            # When user input is submitted, transform, and store it in the chat state
            # (and make sure this runs before other effects since when the user
            #  calls `.messages()`, they should get the latest user input)
            @reactive.effect(priority=9999)
            @reactive.event(self._user_input)
            async def _on_user_input():
                msg = ChatMessage(content=self._user_input(), role="user")
                # It's possible that during the transform, a message is appended, so get
                # the length now, so we can insert the new message at the right index
                n_pre = len(self._messages())
                msg_post = await self._transform_message(msg)
                if msg_post is not None:
                    self._store_message(msg_post)
                    self._suspend_input_handler = False
                else:
                    # A transformed value of None is a special signal to suspend input
                    # handling (i.e., don't generate a response)
                    self._store_message(as_transformed_message(msg), index=n_pre)
                    await self._remove_loading_message()
                    self._suspend_input_handler = True

            self._effects.append(_init_chat)
            self._effects.append(_on_user_input)

        # Prevent repeated calls to Chat() with the same id from accumulating effects
        instance_id = self.id + "_session" + self._session.id
        instance = CHAT_INSTANCES.pop(instance_id, None)
        if instance is not None:
            instance.destroy()
        CHAT_INSTANCES[instance_id] = self

    def ui(
        self,
        *,
        messages: Optional[Sequence[str | ChatMessage]] = None,
        placeholder: str = "Enter a message...",
        width: CssUnit = "min(680px, 100%)",
        height: CssUnit = "auto",
        fill: bool = True,
        **kwargs: TagAttrValue,
    ) -> Tag:
        """
        Place a chat component in the UI.

        This method is only relevant fpr Shiny Express. In Shiny Core, use
        :func:`~shiny.ui.chat_ui` instead to insert the chat UI.

        Parameters
        ----------
        messages
            A sequence of messages to display in the chat. Each message can be either a
            string or a dictionary with `content` and `role` keys. The `content` key
            should contain the message text, and the `role` key can be "assistant" or
            "user".
        placeholder
            Placeholder text for the chat input.
        width
            The width of the chat container.
        height
            The height of the chat container.
        fill
            Whether the chat should vertically take available space inside a fillable
            container.
        kwargs
            Additional attributes for the chat container element.
        """
        return chat_ui(
            id=self.id,
            messages=messages,
            placeholder=placeholder,
            width=width,
            height=height,
            fill=fill,
            **kwargs,
        )

    @overload
    def on_user_submit(
        self, fn: SubmitFunction | SubmitFunctionAsync
    ) -> reactive.Effect_: ...

    @overload
    def on_user_submit(
        self,
    ) -> Callable[[SubmitFunction | SubmitFunctionAsync], reactive.Effect_]: ...

    def on_user_submit(
        self, fn: SubmitFunction | SubmitFunctionAsync | None = None
    ) -> (
        reactive.Effect_
        | Callable[[SubmitFunction | SubmitFunctionAsync], reactive.Effect_]
    ):
        """
        Define a function to invoke when user input is submitted.

        Apply this method as a decorator to a function (`fn`) that should be invoked when the
        user submits a message. The function should take no arguments.

        In many cases, the implementation of `fn` should do at least the following:

        1. Call `.messages()` to obtain the current chat history.
        2. Generate a response based on those messages.
        3. Append the response to the chat history using `.append_message()` (
           or `.append_message_stream()` if the response is streamed).

        Parameters
        ----------
        fn
            A function to invoke when user input is submitted.

        Note
        ----
        This method creates a reactive effect that only gets invalidated when the user
        submits a message. Thus, the function `fn` can read other reactive dependencies,
        but it will only be re-invoked when the user submits a message.
        """

        def create_effect(fn: SubmitFunction | SubmitFunctionAsync):
            afunc = _utils.wrap_async(fn)

            @reactive.effect
            @reactive.event(self._user_input)
            async def handle_user_input():
                if self._suspend_input_handler:
                    from .. import req

                    req(False)
                try:
                    await afunc()
                except Exception as e:
                    await self._raise_exception(e)

            self._effects.append(handle_user_input)

            return handle_user_input

        if fn is None:
            return create_effect
        else:
            return create_effect(fn)

    async def _raise_exception(
        self,
        e: BaseException,
    ) -> None:
        if self.on_error == "unhandled":
            raise e
        else:
            await self._remove_loading_message()
            sanitize = self.on_error == "sanitize"
            raise NotifyException(str(e), sanitize=sanitize) from e

    @overload
    def messages(
        self,
        *,
        format: Literal["anthropic"] = "anthropic",
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[AnthropicMessage, ...]: ...

    @overload
    def messages(
        self,
        *,
        format: Literal["google"] = "google",
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[GoogleMessage, ...]: ...

    @overload
    def messages(
        self,
        *,
        format: Literal["langchain"] = "langchain",
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[LangChainMessage, ...]: ...

    @overload
    def messages(
        self,
        *,
        format: Literal["openai"] = "openai",
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[OpenAIMessage, ...]: ...

    @overload
    def messages(
        self,
        *,
        format: Literal["ollama"] = "ollama",
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[OllamaMessage, ...]: ...

    @overload
    def messages(
        self,
        *,
        format: MISSING_TYPE = MISSING,
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[ChatMessage, ...]: ...

    def messages(
        self,
        *,
        format: MISSING_TYPE | ProviderMessageFormat = MISSING,
        token_limits: tuple[int, int] | None = None,
        transform_user: Literal["all", "last", "none"] = "all",
        transform_assistant: bool = False,
    ) -> tuple[ChatMessage | ProviderMessage, ...]:
        """
        Reactively read chat messages

        Obtain chat messages within a reactive context. The default behavior is
        intended for passing messages along to a model for response generation where
        you typically want to:

        1. Cap the number of tokens sent in a single request (i.e., `token_limits`).
        2. Apply user input transformations (i.e., `transform_user`), if any.
        3. Not apply assistant response transformations (i.e., `transform_assistant`)
           since these are predominantly for display purposes (i.e., the model shouldn't
           concern itself with how the responses are displayed).

        Parameters
        ----------
        format
            The message format to return. The default value of `MISSING` means
            chat messages are returned as :class:`ChatMessage` objects (a dictionary
            with `content` and `role` keys). Other supported formats include:

            * `"anthropic"`: Anthropic message format.
            * `"google"`: Google message (aka content) format.
            * `"langchain"`: LangChain message format.
            * `"openai"`: OpenAI message format.
            * `"ollama"`: Ollama message format.
        token_limits
            Limit the conversation history based on token limits. If specified, only
            the most recent messages that fit within the token limits are returned. This
            is useful for avoiding "exceeded token limit" errors when sending messages
            to the relevant model, while still providing the most recent context available.
            A specified value must be a tuple of two integers. The first integer is the
            maximum number of tokens that can be sent to the model in a single request.
            The second integer is the amount of tokens to reserve for the model's response.
            Note that token counts based on the `tokenizer` provided to the `Chat`
            constructor.
        transform_user
            Whether to return user input messages with transformation applied. This only
            matters if a `transform_user_input` was provided to the chat constructor.
            The default value of `"all"` means all user input messages are transformed.
            The value of `"last"` means only the last user input message is transformed.
            The value of `"none"` means no user input messages are transformed.
        transform_assistant
            Whether to return assistant messages with transformation applied. This only
            matters if an `transform_assistant_response` was provided to the chat
            constructor.

        Note
        ----
        Messages are listed in the order they were added. As a result, when this method
        is called in a `.on_user_submit()` callback (as it most often is), the last
        message will be the most recent one submitted by the user.

        Returns
        -------
        tuple[ChatMessage, ...]
            A tuple of chat messages.
        """

        messages = self._messages()

        # Anthropic requires a user message first and no system messages
        if format == "anthropic":
            messages = self._trim_anthropic_messages(messages)

        if token_limits is not None:
            messages = self._trim_messages(messages, token_limits, format)

        res: list[ChatMessage | ProviderMessage] = []
        for i, m in enumerate(messages):
            transform = False
            if m["role"] == "assistant":
                transform = transform_assistant
            elif m["role"] == "user":
                transform = transform_user == "all" or (
                    transform_user == "last" and i == len(messages) - 1
                )
            content_key = m["transform_key" if transform else "pre_transform_key"]
            chat_msg = ChatMessage(content=str(m[content_key]), role=m["role"])
            if not isinstance(format, MISSING_TYPE):
                chat_msg = as_provider_message(chat_msg, format)
            res.append(chat_msg)

        return tuple(res)

    async def append_message(self, message: Any) -> None:
        """
        Append a message to the chat.

        Parameters
        ----------
        message
            The message to append. A variety of message formats are supported including
            a string, a dictionary with `content` and `role` keys, or a relevant chat
            completion object from platforms like OpenAI, Anthropic, Ollama, and others.

        Note
        ----
        Use `.append_message_stream()` instead of this method when `stream=True` (or
        similar) is specified in model's completion method.
        """
        await self._append_message(message)

    async def _append_message(
        self, message: Any, *, chunk: ChunkOption = False, stream_id: str | None = None
    ) -> None:
        # If currently we're in a stream, handle other messages (outside the stream) later
        if not self._can_append_message(stream_id):
            self._pending_messages.append((message, chunk, stream_id))
            return

        # Update current stream state
        self._current_stream_id = stream_id
        if chunk == "end":
            self._current_stream_id = None

        if chunk is False:
            msg = normalize_message(message)
            chunk_content = None
        else:
            msg = normalize_message_chunk(message)
            # Update the current stream message
            chunk_content = msg["content"]
            self._current_stream_message += chunk_content
            msg["content"] = self._current_stream_message
            if chunk == "end":
                self._current_stream_message = ""

        msg = await self._transform_message(
            msg, chunk=chunk, chunk_content=chunk_content
        )
        if msg is None:
            return
        self._store_message(msg, chunk=chunk)
        await self._send_append_message(msg, chunk=chunk)

    async def append_message_stream(self, message: Iterable[Any] | AsyncIterable[Any]):
        """
        Append a message as a stream of message chunks.

        Parameters
        ----------
        message
            An iterable or async iterable of message chunks to append. A variety of
            message chunk formats are supported, including a string, a dictionary with
            `content` and `role` keys, or a relevant chat completion object from
            platforms like OpenAI, Anthropic, Ollama, and others.

        Note
        ----
        Use this method (over `.append_message()`) when `stream=True` (or similar) is
        specified in model's completion method.
        """

        message = _utils.wrap_async_iterable(message)

        # Run the stream in the background to get non-blocking behavior
        @reactive.extended_task
        async def _stream_task():
            await self._append_message_stream(message)

        _stream_task()

        # Since the task runs in the background (outside/beyond the current context,
        # if any), we need to manually raise any exceptions that occur
        @reactive.effect
        async def _handle_error():
            e = _stream_task.error()
            if e:
                await self._raise_exception(e)
            _handle_error.destroy()  # type: ignore

    async def _append_message_stream(self, message: AsyncIterable[Any]):
        id = _utils.private_random_id()

        empty = ChatMessage(content="", role="assistant")
        await self._append_message(empty, chunk="start", stream_id=id)

        try:
            async for msg in message:
                await self._append_message(msg, chunk=True, stream_id=id)
        finally:
            await self._append_message(empty, chunk="end", stream_id=id)
            await self._flush_pending_messages()

    async def _flush_pending_messages(self):
        still_pending: list[PendingMessage] = []
        for msg, chunk, stream_id in self._pending_messages:
            if self._can_append_message(stream_id):
                await self._append_message(msg, chunk=chunk, stream_id=stream_id)
            else:
                still_pending.append((msg, chunk, stream_id))
        self._pending_messages = still_pending

    def _can_append_message(self, stream_id: str | None) -> bool:
        if self._current_stream_id is None:
            return True
        return self._current_stream_id == stream_id

    # Send a message to the UI
    async def _send_append_message(
        self,
        message: TransformedMessage,
        chunk: ChunkOption = False,
    ):
        if message["role"] == "system":
            # System messages are not displayed in the UI
            return

        if chunk:
            msg_type = "shiny-chat-append-message-chunk"
        else:
            msg_type = "shiny-chat-append-message"

        chunk_type = None
        if chunk == "start":
            chunk_type = "message_start"
        elif chunk == "end":
            chunk_type = "message_end"

        content = message["content_client"]
        content_type = "html" if isinstance(content, HTML) else "markdown"

        msg = ClientMessage(
            content=str(content),
            role=message["role"],
            content_type=content_type,
            chunk_type=chunk_type,
        )

        # print(msg)

        await self._send_custom_message(msg_type, msg)
        # TODO: Joe said it's a good idea to yield here, but I'm not sure why?
        # await asyncio.sleep(0)

    @overload
    def transform_user_input(
        self, fn: TransformUserInput | TransformUserInputAsync
    ) -> None: ...

    @overload
    def transform_user_input(
        self,
    ) -> Callable[[TransformUserInput | TransformUserInputAsync], None]: ...

    def transform_user_input(
        self, fn: TransformUserInput | TransformUserInputAsync | None = None
    ) -> None | Callable[[TransformUserInput | TransformUserInputAsync], None]:
        """
        Transform user input.

        Use this method as a decorator on a function (`fn`) that transforms user input
        before storing it in the chat messages returned by `.messages()`. This is
        useful for implementing RAG workflows, like taking a URL and scraping it for
        text before sending it to the model.

        Parameters
        ----------
        fn
            A function to transform user input before storing it in the chat
            `.messages()`. If `fn` returns `None`, the user input is effectively
            ignored, and `.on_user_submit()` callbacks are suspended until more input is
            submitted. This behavior is often useful to catch and handle errors that
            occur during transformation. In this case, the transform function should
            append an error message to the chat (via `.append_message()`) to inform the
            user of the error.
        """

        def _set_transform(fn: TransformUserInput | TransformUserInputAsync):
            self._transform_user = _utils.wrap_async(fn)

        if fn is None:
            return _set_transform
        else:
            return _set_transform(fn)

    @overload
    def transform_assistant_response(
        self, fn: TransformAssistantResponseFunction
    ) -> None: ...

    @overload
    def transform_assistant_response(
        self,
    ) -> Callable[[TransformAssistantResponseFunction], None]: ...

    def transform_assistant_response(
        self,
        fn: TransformAssistantResponseFunction | None = None,
    ) -> None | Callable[[TransformAssistantResponseFunction], None]:
        """
        Transform assistant responses.

        Use this method as a decorator on a function (`fn`) that transforms assistant
        responses before displaying them in the chat. This is useful for post-processing
        model responses before displaying them to the user.

        Parameters
        ----------
        fn
            A function that takes a string and returns either a string,
            :class:`shiny.ui.HTML`, or `None`. If `fn` returns a string, it gets
            interpreted and parsed as a markdown on the client (and the resulting HTML
            is then sanitized). If `fn` returns :class:`shiny.ui.HTML`, it will be
            displayed as-is. If `fn` returns `None`, the response is effectively ignored.

        Note
        ----
        When doing an `.append_message_stream()`, `fn` gets called on every chunk of the
        response (thus, it should be performant), and can optionally access more
        information (i.e., arguments) about the stream. The 1st argument (required)
        contains the accumulated content, the 2nd argument (optional) contains the
        current chunk, and the 3rd argument (optional) is a boolean indicating whether
        this chunk is the last one in the stream.
        """

        def _set_transform(
            fn: TransformAssistantResponseFunction,
        ):
            nparams = len(inspect.signature(fn).parameters)
            if nparams == 1:
                fn = cast(
                    Union[TransformAssistantResponse, TransformAssistantResponseAsync],
                    fn,
                )
                fn = _utils.wrap_async(fn)

                async def _transform_wrapper(content: str, chunk: str, done: bool):
                    return await fn(content)

                self._transform_assistant = _transform_wrapper

            elif nparams == 3:
                fn = cast(
                    Union[
                        TransformAssistantResponseChunk,
                        TransformAssistantResponseChunkAsync,
                    ],
                    fn,
                )
                self._transform_assistant = _utils.wrap_async(fn)
            else:
                raise Exception(
                    "A @transform_assistant_response function must take 1 or 3 arguments"
                )

        if fn is None:
            return _set_transform
        else:
            return _set_transform(fn)

    async def _transform_message(
        self,
        message: ChatMessage,
        chunk: ChunkOption = False,
        chunk_content: str | None = None,
    ) -> TransformedMessage | None:

        res = as_transformed_message(message)
        key = res["transform_key"]

        if message["role"] == "user" and self._transform_user is not None:
            content = await self._transform_user(message["content"])

        elif message["role"] == "assistant" and self._transform_assistant is not None:
            content = await self._transform_assistant(
                message["content"],
                chunk_content or "",
                chunk == "end" or chunk is False,
            )
        else:
            return res

        if content is None:
            return None

        res[key] = content  # type: ignore

        return res

    # Just before storing, handle chunk msg type and calculate tokens
    def _store_message(
        self,
        message: TransformedMessage,
        chunk: ChunkOption = False,
        index: int | None = None,
    ) -> None:

        # Don't actually store chunks until the end
        if chunk is True or chunk == "start":
            return None

        with reactive.isolate():
            messages = self._messages()

        if index is None:
            index = len(messages)

        messages = list(messages)
        messages.insert(index, message)

        self._messages.set(tuple(messages))
        if message["role"] == "user":
            self._latest_user_input.set(message)

        return None

    def _trim_messages(
        self,
        messages: tuple[TransformedMessage, ...],
        token_limits: tuple[int, int],
        format: MISSING_TYPE | ProviderMessageFormat,
    ) -> tuple[TransformedMessage, ...]:

        n_total, n_reserve = token_limits
        if n_total <= n_reserve:
            raise ValueError(
                f"Invalid token limits: {token_limits}. The 1st value must be greater "
                "than the 2nd value."
            )

        # Since don't trim system messages, 1st obtain their total token count
        # (so we can determine how many non-system messages can fit)
        n_system_tokens: int = 0
        n_system_messages: int = 0
        n_other_messages: int = 0
        token_counts: list[int] = []
        for m in messages:
            count = self._get_token_count(m["content_server"])
            token_counts.append(count)
            if m["role"] == "system":
                n_system_tokens += count
                n_system_messages += 1
            else:
                n_other_messages += 1

        remaining_non_system_tokens = n_total - n_reserve - n_system_tokens

        if remaining_non_system_tokens <= 0:
            raise ValueError(
                f"System messages exceed `.messages(token_limits={token_limits})`. "
                "Consider increasing the 1st value of `token_limit` or setting it to "
                "`token_limit=None` to disable token limits."
            )

        # Now, iterate through the messages in reverse order and appending
        # until we run out of tokens
        messages2: list[TransformedMessage] = []
        n_other_messages2: int = 0
        token_counts.reverse()
        for i, m in enumerate(reversed(messages)):
            if m["role"] == "system":
                messages2.append(m)
                continue
            remaining_non_system_tokens -= token_counts[i]
            if remaining_non_system_tokens >= 0:
                messages2.append(m)
                n_other_messages2 += 1

        messages2.reverse()

        if len(messages2) == n_system_messages and n_other_messages2 > 0:
            raise ValueError(
                f"Only system messages fit within `.messages(token_limits={token_limits})`. "
                "Consider increasing the 1st value of `token_limit` or setting it to "
                "`token_limit=None` to disable token limits."
            )

        return tuple(messages2)

    def _trim_anthropic_messages(
        self,
        messages: tuple[TransformedMessage, ...],
    ) -> tuple[TransformedMessage, ...]:

        if any(m["role"] == "system" for m in messages):
            raise ValueError(
                "Anthropic requires a system prompt to be specified in it's `.create()` method "
                "(not in the chat messages with `role: system`)."
            )
        for i, m in enumerate(messages):
            if m["role"] == "user":
                return messages[i:]

        return ()

    def _get_token_count(
        self,
        content: str,
    ) -> int:
        if self._tokenizer is None:
            self._tokenizer = get_default_tokenizer()

        if self._tokenizer is None:
            raise ValueError(
                "A tokenizer is required to impose `token_limits` on messages. "
                "To get a generic default tokenizer, install the `tokenizers` "
                "package (`pip install tokenizers`). "
                "To get a more precise token count, provide a specific tokenizer "
                "to the `Chat` constructor."
            )

        encoded = self._tokenizer.encode(content)
        if isinstance(encoded, TokenizersEncoding):
            return len(encoded.ids)
        else:
            return len(encoded)

    def user_input(self, transform: bool = False) -> str | None:
        """
        Reactively read the user's message.

        Parameters
        ----------
        transform
            Whether to apply the user input transformation function (if one was
            provided).

        Returns
        -------
        str | None
            The user input message (before any transformation).

        Note
        ----
        Most users shouldn't need to use this method directly since the last item in
        `.messages()` contains the most recent user input. It can be useful for:

          1. Taking a reactive dependency on the user's input outside of a `.on_user_submit()` callback.
          2. Maintaining message state separately from `.messages()`.

        """
        msg = self._latest_user_input()
        if msg is None:
            return None
        key = "content_server" if transform else "content_client"
        return str(msg[key])

    def _user_input(self) -> str:
        id = self.user_input_id
        return cast(str, self._session.input[id]())

    def update_user_input(
        self, *, value: str | None = None, placeholder: str | None = None
    ):
        """
        Update the user input.

        Parameters
        ----------
        value
            The value to set the user input to.
        placeholder
            The placeholder text for the user input.
        """

        obj = _utils.drop_none({"value": value, "placeholder": placeholder})

        _utils.run_coro_sync(
            self._session.send_custom_message(
                "shinyChatMessage",
                {
                    "id": self.id,
                    "handler": "shiny-chat-update-user-input",
                    "obj": obj,
                },
            )
        )

    def set_user_message(self, value: str):
        """
        Deprecated. Use `update_user_input(value=value)` instead.
        """

        warn_deprecated(
            "set_user_message() is deprecated. Use update_user_input(value=value) instead."
        )

        self.update_user_input(value=value)

    async def clear_messages(self):
        """
        Clear all chat messages.
        """
        self._messages.set(())
        await self._send_custom_message("shiny-chat-clear-messages", None)

    def destroy(self):
        """
        Destroy the chat instance.
        """
        for x in self._effects:
            x.destroy()
        self._effects.clear()

    async def _remove_loading_message(self):
        await self._send_custom_message("shiny-chat-remove-loading-message", None)

    async def _send_custom_message(self, handler: str, obj: ClientMessage | None):
        await self._session.send_custom_message(
            "shinyChatMessage",
            {
                "id": self.id,
                "handler": handler,
                "obj": obj,
            },
        )


@add_example(ex_dir="../api-examples/chat")
def chat_ui(
    id: str,
    *,
    messages: Optional[Sequence[str | ChatMessage]] = None,
    placeholder: str = "Enter a message...",
    width: CssUnit = "min(680px, 100%)",
    height: CssUnit = "auto",
    fill: bool = True,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    UI container for a chat component (Shiny Core).

    This function is for locating a :class:`~shiny.ui.Chat` instance in a Shiny Core
    app. If you are using Shiny Express, use the :method:`~shiny.ui.Chat.ui` method
    instead.

    Parameters
    ----------
    id
        A unique identifier for the chat UI.
    messages
        A sequence of messages to display in the chat. Each message can be either a string
        or a dictionary with a `content` and `role` key. The `content` key should contain
        the message text, and the `role` key can be "assistant" or "user".
    placeholder
        Placeholder text for the chat input.
    width
        The width of the chat container.
    height
        The height of the chat container.
    fill
        Whether the chat should vertically take available space inside a fillable container.
    kwargs
        Additional attributes for the chat container element.
    """

    id = resolve_id(id)

    message_tags: list[Tag] = []
    if messages is None:
        messages = []
    for msg in messages:
        if isinstance(msg, str):
            msg = {"content": msg, "role": "assistant"}
        elif isinstance(msg, dict):
            if "content" not in msg:
                raise ValueError("Each message must have a 'content' key.")
            if "role" not in msg:
                raise ValueError("Each message must have a 'role' key.")
        else:
            raise ValueError("Each message must be a string or a dictionary.")

        if msg["role"] == "user":
            tag_name = "shiny-user-message"
        else:
            tag_name = "shiny-chat-message"

        message_tags.append(Tag(tag_name, content=msg["content"]))

    res = Tag(
        "shiny-chat-container",
        Tag("shiny-chat-messages", *message_tags),
        Tag(
            "shiny-chat-input",
            id=f"{id}_user_input",
            placeholder=placeholder,
        ),
        chat_deps(),
        {
            "style": css(
                width=as_css_unit(width),
                height=as_css_unit(height),
            )
        },
        id=id,
        placeholder=placeholder,
        fill=fill,
        **kwargs,
    )

    if fill:
        res = as_fillable_container(as_fill_item(res))

    return res


def as_transformed_message(message: ChatMessage) -> TransformedMessage:
    if message["role"] == "user":
        transform_key = "content_server"
        pre_transform_key = "content_client"
    else:
        transform_key = "content_client"
        pre_transform_key = "content_server"

    return TransformedMessage(
        content_client=message["content"],
        content_server=message["content"],
        role=message["role"],
        transform_key=transform_key,
        pre_transform_key=pre_transform_key,
    )


CHAT_INSTANCES: WeakValueDictionary[str, Chat] = WeakValueDictionary()



from datetime import date

from shiny.express import ui

# Default start and end is the current date in the client's time zone
ui.input_date_range("daterange1", "Date range:")
# Set start and end dates
ui.input_date_range(
    "daterange2", "Set start and end date:", start="2001-01-01", end="2010-12-31"
)
# Start and end are always specified in yyyy-mm-dd, even if the display
# format is different
ui.input_date_range(
    "daterange3",
    "Min, max, start, and end dates are set with custom format and separator:",
    start="2001-01-01",
    end="2010-12-31",
    min="2001-01-01",
    max="2012-12-21",
    format="mm/dd/yy",
    separator=" - ",
)
# Pass in Date objects
ui.input_date_range(
    "daterange4",
    "Default start and end use date objects:",
    start=date(2001, 1, 1),
    end=date(2010, 12, 31),
)
# Use different language and different first day of week
ui.input_date_range(
    "daterange5",
    "Language is German and we starts on Monday:",
    language="de",
    weekstart=1,
)
# Start with decade view instead of default month view
ui.input_date_range(
    "daterange6", "Start Date picker in Decade view:", startview="decade"
)
# Set width of the daterange field
ui.input_date_range("daterange7", "Set width of text input:", width="600px")
# Set autoclose to false
ui.input_date_range("daterange8", "Auto close is disabled:", autoclose=False)



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("output_ui")


def test_output_ui_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    more_controls = controller.OutputUi(page, "moreControls")

    more_controls.expect_inline(False)

    more_controls.expect_empty(True)
    expect(more_controls.loc).to_have_text("")
    expect(page.locator("#n")).to_have_count(0)
    expect(page.locator("#label")).to_have_count(0)

    controller.InputActionButton(page, "add").click()

    more_controls.expect_empty(False)
    expect(more_controls.loc).not_to_have_text("")
    expect(page.locator("#n")).to_have_count(1)
    expect(page.locator("#label")).to_have_count(1)

    controller.InputSlider(page, "n").expect_value("500")
    controller.InputText(page, "label").expect_value("")



from shiny.express import input, render, ui

ui.input_radio_buttons(
    "rb",
    "Choose one:",
    {
        "html": ui.HTML("<span style='color:red;'>Red Text</span>"),
        "text": "Normal text",
    },
)


@render.express
def val():
    "You chose " + input.rb()



from shiny import reactive
from shiny.express import input, ui


def run_model(delay=10.0):
    import time

    # Pretend to run a model for `delay` seconds
    start_time = time.time()
    while time.time() - start_time < delay:
        pass
    return time.time()


ui.input_action_button("run", "Run Model")

model_result = reactive.value()


def the_modal():
    return ui.modal(
        "The model is running, please wait.",
        title="Running model",
        easy_close=False,
        footer=None,
    )


@reactive.effect
@reactive.event(input.run)
def do_run_model():
    # Show the modal, blocking interaction with the UI
    ui.modal_show(the_modal())

    result = run_model(delay=4)

    # Now that we have model results, remove the modal
    # and update the model result reactive value
    ui.modal_remove()
    model_result.set(result)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_page_navbar_sidebar(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    page_navbar_sidebar = controller.PageNavbar(page, "page_navbar_sidebar")
    page_navbar_sidebar.expect_value("Data")
    page_navbar_sidebar._expect_content_text(
        "This page could be used to pick a dataset."
    )
    page_navbar_sidebar.expect_sidebar(True)



from icons import question_circle_fill

from shiny.express import ui

with ui.tooltip(id="btn_tooltip"):
    ui.input_action_button("btn", "A button", class_="mt-3")

    "A message"

with ui.card(class_="mt-3"):
    with ui.card_header():
        with ui.tooltip(placement="right", id="card_tooltip"):
            ui.span("Card title ", question_circle_fill)
            "Additional info"

    "Card body content..."



from shiny import App, ui

y = ui.card("A simple card")

app_ui = ui.page_fluid(
    # Always has 2 columns (on non-mobile)
    ui.layout_column_wrap(y, y, y, width=1 / 2),
    ui.hr(),
    # Has three columns when viewport is wider than 750px
    ui.layout_column_wrap(y, y, y, width="250px"),
)


app = App(app_ui, server=None)



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Optional

from htmltools import (
    Tag,
    TagAttrs,
    TagAttrValue,
    TagChild,
    TagFunction,
    Tagifiable,
    css,
    tags,
)

from .._docstring import add_example, no_example
from ._card import CardItem, card, card_body
from ._tag import consolidate_attrs
from ._utils import css_no_sub
from .css import as_css_unit
from .css._css_unit import CssUnit, as_grid_unit
from .fill import as_fill_item, as_fillable_container

__all__ = (
    "value_box",
    "showcase_left_center",
    "showcase_top_right",
    "showcase_bottom",
    "value_box_theme",
    "ValueBoxTheme",
    "ShowcaseLayout",
)

# TODO-future; Better documentation to bslib's value box and theme;
# Related: https://rstudio.github.io/bslib/reference/value_box.html#themes,
# Related: https://github.com/rstudio/bslib/pull/758/files

# Define literal types
SHOWCASE_LAYOUTS_STR = Literal["left center", "top right", "bottom"]
SHOWCASE_LAYOUTS_VALS = ("left center", "top right", "bottom")


@dataclass
class ShowcaseLayout:
    """
    Showcase layout

    Base layout information utilized to display :func:`~shiny.ui.value_box`'s `showcase` value.

    See Also
    --------
    * :func:`~shiny.ui.showcase_left_center`
    * :func:`~shiny.ui.showcase_top_right`
    * :func:`~shiny.ui.showcase_bottom`
    * :func:`~shiny.ui.value_box`
    """

    class_: str
    """CSS class to set on the layout"""
    width: CssUnit | None
    """Width of the showcase area"""
    width_full_screen: CssUnit | None
    """Width of the showcase area when the value box is full screen"""
    height: CssUnit | None
    """Height of the showcase area"""
    height_full_screen: CssUnit | None
    """Height of the showcase area when the value box is full screen"""
    max_height: CssUnit | None
    """Maximum height of the showcase area"""
    max_height_full_screen: CssUnit | None
    """Maximum height of the showcase area when the value box is full screen"""

    def __init__(
        self,
        *,
        class_: str,
        width: CssUnit | None = "33%",
        width_full_screen: CssUnit | None = "1fr",
        height: CssUnit | None = None,
        height_full_screen: CssUnit | None = None,
        max_height: CssUnit | None = "100px",
        max_height_full_screen: CssUnit | None = "67%",
    ) -> None:
        self.class_ = class_
        self.width = as_grid_unit(width)
        self.width_full_screen = as_grid_unit(width_full_screen)
        self.height = as_grid_unit(height)
        self.height_full_screen = as_grid_unit(height_full_screen)
        self.max_height = as_css_unit(max_height)
        self.max_height_full_screen = as_css_unit(max_height_full_screen)


@add_example()
def showcase_left_center(
    *,
    width: CssUnit = "30%",
    width_full_screen: CssUnit | None = "1fr",
    max_height: CssUnit | None = "100px",
    max_height_full_screen: CssUnit | None = "67%",
) -> ShowcaseLayout:
    """
    Showcase left center

    A :func:`~shiny.ui.showcase_left_center` is a `ShowcaseLayout` with
    the following default properties:

    * `width` is `"30%"`
    * `width_full_screen` is `"1fr"`
    * `max_height` is `"100px"`
    * `max_height_full_screen` is `"67%"`

    See Also
    --------
    * :func:`~shiny.ui.showcase_top_right`
    * :func:`~shiny.ui.showcase_bottom`
    * :func:`~shiny.ui.value_box`
    """
    return ShowcaseLayout(
        class_="showcase-left-center",
        width=width,
        width_full_screen=width_full_screen,
        max_height=max_height,
        max_height_full_screen=max_height_full_screen,
    )


@add_example()
def showcase_top_right(
    *,
    width: CssUnit = "40%",
    width_full_screen: CssUnit | None = "1fr",
    max_height: CssUnit | None = "75px",
    max_height_full_screen: CssUnit | None = "67%",
) -> ShowcaseLayout:
    """
    Showcase top right

    A :func:`~shiny.ui.showcase_top_right` is a `ShowcaseLayout` with
    the following default properties:

    * `width` is `"40%"`
    * `width_full_screen` is `"1fr"`
    * `max_height` is `"75px"`
    * `max_height_full_screen` is `"67%"`

    See Also
    --------
    * :func:`~shiny.ui.showcase_left_center`
    * :func:`~shiny.ui.showcase_bottom`
    * :func:`~shiny.ui.value_box`
    """

    return ShowcaseLayout(
        class_="showcase-top-right",
        width=width,
        width_full_screen=width_full_screen,
        max_height=max_height,
        max_height_full_screen=max_height_full_screen,
    )


@add_example()
def showcase_bottom(
    *,
    width: CssUnit = "100%",
    width_full_screen: CssUnit | None = None,
    height: CssUnit | None = "auto",
    height_full_screen: CssUnit | None = "2fr",
    max_height: CssUnit | None = "100px",
    max_height_full_screen: CssUnit | None = None,
) -> ShowcaseLayout:
    """
    Showcase bottom

    A :func:`~shiny.ui.showcase_bottom` is a `ShowcaseLayout` with
    the following default properties:

    * `width` is `"100%"`
    * `width_full_screen` is `None`
    * `height` is `"auto"`
    * `height_full_screen` is `"2fr"`
    * `max_height` is `"100px"`
    * `max_height_full_screen` is `None`

    See Also
    --------
    * :func:`~shiny.ui.showcase_left_center`
    * :func:`~shiny.ui.showcase_top_right`
    * :func:`~shiny.ui.value_box`
    """

    return ShowcaseLayout(
        class_="showcase-bottom",
        width=width,
        width_full_screen=width_full_screen,
        height=height,
        height_full_screen=height_full_screen,
        max_height=max_height,
        max_height_full_screen=max_height_full_screen,
    )


def resolve_showcase_layout(
    showcase_layout: SHOWCASE_LAYOUTS_STR | ShowcaseLayout,
) -> ShowcaseLayout:
    if isinstance(showcase_layout, ShowcaseLayout):
        return showcase_layout

    if showcase_layout == "left center":
        return showcase_left_center()
    elif showcase_layout == "top right":
        return showcase_top_right()
    elif showcase_layout == "bottom":
        return showcase_bottom()
    else:
        showcase_layouts = "'" + "', '".join(SHOWCASE_LAYOUTS_VALS) + "'"
        raise ValueError(
            f"`showcase_layout` must be one of {showcase_layouts} or inherit from `ShowcaseLayout`"
        )


@dataclass
class ValueBoxTheme:
    class_: str | None
    fg: str | None
    bg: str | None


@no_example()
def value_box_theme(
    name: Optional[str] = None,
    *,
    fg: Optional[str] = None,
    bg: Optional[str] = None,
) -> ValueBoxTheme:
    """
    Value box theme

    A theme for a :func:`~shiny.ui.value_box`. Themes provide a convenient way to use
    your app's Bootstrap theme colors as the foreground or background colors of the
    value box. For more control, you can create your own theme with
    :func:`~shiny.ui.value_box_theme` where you can pass foreground and background
    value.

    See
    [rstudio/bslib#themes](https://rstudio.github.io/bslib/reference/value_box.html#themes)
    for more examples.


    Parameters
    ---------
    name
        The name of the theme, e.g. `"primary"`, `"danger"`, `"purple"`.  `name` can
        also be a Bootstrap-supported color: `"blue"`, `"purple"`, `"pink"`, `"red"`,
        `"orange"`, `"yellow"`, `"green"`, `"teal"`, and `"cyan"`. These colors can be
        used with `bg-NAME`, `text-NAME`, and `bg-gradient-NAME1-NAME2`. If a `name`
        does not start with `text-` or `bg-`, it will be auto-prefixed with `bg-`.
    fg,bg
        The background and foreground colors for the theme.

    Returns
    -------
    :
        A `ValueBoxTheme`

    See Also
    --------
    * :func:`~shiny.ui.value_box`
    """
    # bg
    #     If only `bg` is provided,
    #     then the foreground color is automatically chosen from `$black` or `$white` to
    #     provide the best contrast with the background color.
    if name is None:
        if bg is None:
            name = "default"
        else:
            # TODO-future; color contrast
            # # Don't warn if we can't get a contrast color, `bg` might be valid
            # # CSS but not something sass can compute on
            # if fg is None:
            #     fg = get_contrast_color(bg, warn=False)
            ...
        return ValueBoxTheme(class_=name, bg=bg, fg=fg)

    if not isinstance(name, str):
        raise TypeError(
            """`value_box_theme(theme=)` should be a single string, """
            """e.g. `"primary"`, `"danger"`, `"purple"`, etc."""
        )

    if not (name.startswith("bg-") or name.startswith("text-")):
        name = "bg-" + name

    return ValueBoxTheme(class_=name, bg=bg, fg=fg)


@add_example()
def value_box(
    title: TagChild,
    value: TagChild,
    *args: TagChild | TagAttrs,
    showcase: Optional[TagChild] = None,
    showcase_layout: SHOWCASE_LAYOUTS_STR | ShowcaseLayout = "left center",
    full_screen: bool = False,
    theme: Optional[str | ValueBoxTheme] = None,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    id: Optional[str] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Value box

    An opinionated (:func:`~shiny.ui.card`-powered) box, designed for
    displaying a `value` and `title`. Optionally, a `showcase` can provide context
    for what the `value` represents (for example, it could hold an icon, or even a
    :func:`~shiny.ui.output_plot`).

    Parameters
    ----------
    title,value
        A string, number, or :class:`~htmltools.Tag` child to display as
        the title or value of the value box. The `title` appears above the `value`.
    *args
        Unnamed arguments may be any :class:`~htmltools.Tag` children to display below
        `value`. Named arguments are passed to :func:`~shiny.ui.card` as
        element attributes.
    showcase
        A :class:`~htmltools.Tag` child to showcase (e.g., an icon, a
        :func:`~shiny.ui.output_plot`, etc).
    showcase_layout
        One of `"left center"` (default), `"top right"` or `"bottom"`. Alternatively,
        you can customize the showcase layout options with the
        :func:`~shiny.ui.showcase_left_center`, :func:`~shiny.ui.showcase_top_right`,
        or :func:`~shiny.ui.showcase_bottom` functions. Use the options functions when
        you want to control the height or width of the showcase area.
    theme
        The name of a theme (e.g. `"primary"`, `"danger"`, `"purple"`, `"bg-green"`,
        `"text-red"`) for the value box, or a theme constructed with
        :func:`~shiny.ui.value_box_theme`.

        The theme names provide a convenient way to use your app's Bootstrap theme
        colors as the foreground or background colors of the value box. For more
        control, you can create your own theme with :func:`~shiny.ui.value_box_theme`
        where you can pass foreground and background colors directly.

        Bootstrap supported color themes: `"blue"`, `"purple"`, `"pink"`, `"red"`,
        `"orange"`, `"yellow"`, `"green"`, `"teal"`, and `"cyan"`. These colors can be
        used with `bg-NAME`, `text-NAME`, and `bg-gradient-NAME1-NAME2` to change the
        background, foreground, or use a background gradient respectively.

        If a `theme` string does not start with `text-` or `bg-`, it will be auto
        prefixed with `bg-`.
    full_screen
        If `True`, an icon will appear when hovering over the card body. Clicking the
        icon expands the card to fit viewport size.
    height,max_height,min_height
        Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a value box is
        made `full_screen`.
    fill
        Whether to allow the value box to grow/shrink to fit a fillable container with
        an opinionated height (e.g., :func:`~shiny.ui.page_fillable`).
    class_
        Utility classes for customizing the appearance of the summary card. Use `bg-*`
        and `text-*` classes (e.g, `"bg-danger"` and `"text-light"`) to customize the
        background/foreground colors.
    id
        Provide a unique identifier for the :func:`~shiny.ui.value_box()` to report its
        state to Shiny. For example, using `id="my_value_box"`, you can observe the
        value box's full screen state with `input.my_value_box()["full_screen"]`.
    **kwargs
        Additional attributes to pass to :func:`~shiny.ui.card`.

    Returns
    -------
    :
        A :func:`~shiny.ui.card`

    See Also
    --------
    * :func:`~shiny.ui.value_box_theme`
    * :func:`~shiny.ui.showcase_bottom`
    * :func:`~shiny.ui.showcase_left_center`
    * :func:`~shiny.ui.showcase_top_right`
    * :func:`~shiny.ui.card`
    """

    # theme
    #     See the **Themes** section for
    #     more details.
    # height,max_height
    #     Any valid CSS unit (e.g., `height="200px"`). Doesn't apply when a card is made
    #     `full_screen` (in this case, consider setting a `height` in
    #     :func:`~shiny.experimental.ui.card_body`).

    attrs, children = consolidate_attrs(
        *args,
        class_=class_,
        **kwargs,
    )

    # ---- Title and value ----
    if showcase_layout is None:  # pyright: ignore[reportUnnecessaryComparison]
        showcase_layout = SHOWCASE_LAYOUTS_VALS[0]
    if isinstance(title, (str, int, float)):
        title = tags.p(title)
    if isinstance(value, (str, int, float)):
        value = tags.p(value)

    if title is not None:
        title = wrap_in_carrier_tag_with_class("value-box-title", title)
    if value is not None:
        value = wrap_in_carrier_tag_with_class("value-box-value", value)

    # ---- Contents ----
    contents: Tagifiable = wrap_in_carrier_tag_with_class(
        "value-box-area",
        title,
        value,
        *children,
    )

    # ---- Showcase ----
    if showcase is not None:
        showcase_layout = resolve_showcase_layout(showcase_layout)
        contents = render_showcase_layout(
            showcase_layout=showcase_layout,
            showcase=showcase,
            contents=contents,
        )

    if not isinstance(theme, ValueBoxTheme):
        theme = value_box_theme(theme)

    # ---- Layout ----

    return card(
        {
            "class": "bslib-value-box",
            "style": css_no_sub(
                **{
                    "color": theme.fg,
                    "background-color": theme.bg,
                    # These variables are used by the full screen card button
                    "--bslib-color-fg": theme.fg,
                    "--bslib-color-bg": theme.bg,
                },
            ),
        },
        {"class": theme.class_} if theme.class_ else None,
        {"class": class_} if class_ else None,
        (
            {"class": showcase_layout.class_}
            if showcase and isinstance(showcase_layout, ShowcaseLayout)
            else None
        ),
        attrs,
        contents,
        full_screen=full_screen,
        height=height,
        max_height=max_height,
        min_height=min_height,
        fill=fill,
        id=id,
    )


def render_showcase_layout(
    # Requiring named args as the typing is possible to mix up
    *,
    showcase_layout: ShowcaseLayout,
    showcase: TagChild,
    contents: Tag,
) -> CardItem:
    showcase = wrap_in_carrier_tag_with_class("value-box-showcase", showcase)

    grid_props = css_no_sub(
        **{
            "--bslib-grid-height": "auto",
            "--bslib-grid-height-mobile": "auto",
            "---bslib-value-box-showcase-w": showcase_layout.width,
            "---bslib-value-box-showcase-w-fs": showcase_layout.width_full_screen,
            "---bslib-value-box-showcase-h": showcase_layout.height,
            "---bslib-value-box-showcase-h-fs": showcase_layout.height_full_screen,
            "---bslib-value-box-showcase-max-h": showcase_layout.max_height,
            "---bslib-value-box-showcase-max-h-fs": showcase_layout.max_height_full_screen,
        },
    )

    value_box_grid = wrap_in_carrier_tag_with_class(
        "value-box-grid",
        showcase,
        contents,
        fillable=False,
        fill=True,
        style=grid_props,
    )

    return card_body(
        value_box_grid,
        style=css(padding=0),
        fillable=True,
    )


def wrap_in_carrier_tag_with_class(
    class_: str,
    *args: TagChild | TagAttrs,
    tag: TagFunction = tags.div,
    fillable: bool = True,
    fill: bool = True,
    **kwargs: TagAttrValue,
) -> Tag:
    ret = tag({"class": class_}, *args, **kwargs)
    if fill:
        ret = as_fill_item(ret)
    if fillable:
        ret = as_fillable_container(ret)
    return ret



try:
    import pytest  # noqa: F401 # pyright: ignore[reportUnusedImport, reportMissingTypeStubs]
except ImportError:
    raise ImportError(
        "The shiny.pytest module requires the pytest package to be installed."
        " Please install it with this command:"
        "\n\n    pip install pytest"
    )

from ._fixture import ScopeName, create_app_fixture

__all__ = ("create_app_fixture", "ScopeName")



from shiny import App, Inputs, Outputs, Session, render, ui

scenarios = dict(
    out1="The following output should be empty",
    out2='The following output should have the word "One"',
    out3='The following output should have the word "Two"',
    out4='The following output should also have the word "Two"',
)

app_ui = ui.page_fluid(
    [
        ui.p(ui.div(desc), ui.output_text_verbatim(id, placeholder=True))
        for id, desc in scenarios.items()
    ]
)


def server(input: Inputs, output: Outputs, session: Session):
    # When @output(id="out2") is added to the function, it should
    # un-register the implicit @output(id="out1") that @render.text
    # does internally.

    @output(id="out2")
    @render.text
    def out1():
        return "One"

    @render.text
    def out3():
        return "Two"

    # Only implicit registration can be revoked. Since out3 was
    # explicitly registered, registering it again should result
    # in both registrations being active.
    output(out3)
    output(id="out4")(out3)


app = App(app_ui, server)



# App altered from: https://github.com/rstudio/py-shiny/blob/main/shiny/api-examples/data_frame/app.py
import pandas  # noqa: F401 (this line needed for Shinylive to load plotly.express)
import plotly.express as px
import plotly.graph_objs as go
from shinywidgets import output_widget, render_widget

from shiny import App, reactive, render, req, session, ui

# Load the Gapminder dataset
df = px.data.gapminder()

# Prepare a summary DataFrame
summary_df = (
    df.groupby("country")
    .agg(
        {
            "pop": ["min", "max", "mean"],
            "lifeExp": ["min", "max", "mean"],
            "gdpPercap": ["min", "max", "mean"],
        }
    )
    .reset_index()
)

summary_df.columns = ["_".join(col).strip() for col in summary_df.columns.values]
summary_df.rename(columns={"country_": "country"}, inplace=True)

app_ui = ui.page_fillable(
    {"class": "p-3"},
    ui.p(
        ui.strong("Instructions:"),
        " Select one or more countries in the table below to see more information.",
    ),
    ui.layout_column_wrap(
        ui.card(
            ui.output_data_frame("summary_data"),
        ),
        ui.layout_column_wrap(
            ui.card(
                output_widget("country_detail_pop", height="100%"),
            ),
            ui.card(
                output_widget("country_detail_percap", height="100%"),
            ),
            width=1 / 2,
        ),
        width=1,
    ),
)


def server(input, output, session):
    @render.data_frame
    def summary_data():
        return render.DataGrid(
            summary_df.round(2),
            selection_mode="rows",
            width="100%",
            height="100%",
        )

    @reactive.calc
    def filtered_df():
        data_selected = summary_data.data_view(selected=True)
        req(not data_selected.empty)
        countries = data_selected["country"]
        # Filter data for selected countries
        return df[df["country"].isin(countries)]

    @render_widget
    def country_detail_pop():
        # Create the plot
        fig = px.line(
            filtered_df(),
            x="year",
            y="pop",
            color="country",
            title="Population Over Time",
        )
        widget = go.FigureWidget(fig)

        @synchronize_size("country_detail_pop")
        def on_size_changed(width, height):
            widget.layout.width = width
            widget.layout.height = height

        return widget

    @render_widget
    def country_detail_percap():
        # Create the plot
        fig = px.line(
            filtered_df(),
            x="year",
            y="gdpPercap",
            color="country",
            title="GDP per Capita Over Time",
        )
        widget = go.FigureWidget(fig)

        @synchronize_size("country_detail_percap")
        def on_size_changed(width, height):
            widget.layout.width = width
            widget.layout.height = height

        return widget


# This is a hacky workaround to help Plotly plots automatically
# resize to fit their container. In the future we'll have a
# built-in solution for this.
def synchronize_size(output_id):
    def wrapper(func):
        input = session.get_current_session().input

        @reactive.effect
        def size_updater():
            func(
                input[f".clientdata_output_{output_id}_width"](),
                input[f".clientdata_output_{output_id}_height"](),
            )

        # When the output that we're synchronizing to is invalidated,
        # clean up the size_updater Effect.
        reactive.get_current_context().on_invalidate(size_updater.destroy)

        return size_updater

    return wrapper


app = App(app_ui, server)



from ._renderer import (
    AsyncValueFn,
    Jsonifiable,
    Renderer,
    RendererT,
    ValueFn,
)

__all__ = (
    "Renderer",
    "ValueFn",
    "Jsonifiable",
    "AsyncValueFn",
    "RendererT",
)



from __future__ import annotations

import json
from typing import Any, Literal, Optional

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, div, tags

from .._docstring import add_example
from .._namespaces import resolve_id_or_none
from .._utils import drop_none
from ._tag import consolidate_attrs
from ._web_component import web_component


@add_example()
def popover(
    trigger: TagChild,
    *args: TagChild | TagAttrs,
    title: Optional[TagChild] = None,
    id: Optional[str] = None,
    placement: Literal["auto", "top", "right", "bottom", "left"] = "auto",
    options: Optional[dict[str, Any]] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Add a popover to a UI element.

    Display additional information when clicking on a UI element (typically a
    button).

    Parameters
    ----------
    trigger
        The UI element to serve as the popover trigger (typically a
        :func:`~shiny.ui.input_action_button` or similar). If `trigger` renders as
        multiple HTML elements (e.g., it's a :class:`~shiny.ui.TagList`), the last
        HTML element is used for the trigger. If the `trigger` should contain all of
        those elements, wrap the object in a :func:`~shiny.ui.tags.div` or
        :func:`~shiny.ui.tags.span`.
    *args
        UI elements for the popover's body. Character strings are
        automatically escaped unless marked as :class:`~shiny.ui.HTML`.
    title
        A title (header) for the popover.
    id
        A character string. Required to re-actively respond to the visibility of the
        popover (via the `input.<ID>()` value) and/or update the visibility/contents of
        the popover.
    placement
        The placement of the popover relative to its trigger.
    options
        A list of additional
        `options <https://getbootstrap.com/docs/5.3/components/popovers/#options>`_.


    Closing popovers
    ----------------

    In addition to clicking the `close_button`, popovers can be closed by pressing the
    Esc/Space key when the popover (and/or its trigger) is focused.

    Accessibility of Popover Triggers
    ---------------------------------

    Because the user needs to interact with the `trigger` element to see the `popover`,
    it's best practice to use an element that is typically accessible via keyboard
    interactions, like a button or a link.

    If you use a non-interactive element, like a `<span>` or text, `popover()` will
    automatically add the `tabindex="0"` attribute to the trigger element to make sure
    that users can reach the element with the keyboard. This means that in most cases
    you can use any element you want as the trigger.

    One place where it's important to consider the accessibility of the trigger is when
    using an icon without any accompanying text. In these cases, many icon elements are
    created with the assumption that the icon is decorative, which will make it
    inaccessible to users of assistive technologies.

    When using an icon as the primary trigger, ensure that the icon does not have
    `aria-hidden="true"` or `role="presentation"` attributes. Icon packages typically
    provide a way to specify a title for the icon, as well as a way to specify that the
    icon is not decorative. The title should be a short description of the purpose of
    the trigger, rather than a description of the icon itself.

    For example:

    ```python
    icon_title = "Settings"
    def bs_gear_icon(title: str):
        # Enhanced from https://rstudio.github.io/bsicons/ via `bsicons::bs_icon("gear", title = icon_title)`
        return ui.HTML(f'<svg xmlns="http://www.w3.org/2000/svg" viewBox="0 0 16 16" class="bi bi-gear " style="height:1em;width:1em;fill:currentColor;" aria-hidden="true" role="img" ><title>{title}</title><path d="M8 4.754a3.246 3.246 0 1 0 0 6.492 3.246 3.246 0 0 0 0-6.492zM5.754 8a2.246 2.246 0 1 1 4.492 0 2.246 2.246 0 0 1-4.492 0z"></path><path d="M9.796 1.343c-.527-1.79-3.065-1.79-3.592 0l-.094.319a.873.873 0 0 1-1.255.52l-.292-.16c-1.64-.892-3.433.902-2.54 2.541l.159.292a.873.873 0 0 1-.52 1.255l-.319.094c-1.79.527-1.79 3.065 0 3.592l.319.094a.873.873 0 0 1 .52 1.255l-.16.292c-.892 1.64.901 3.434 2.541 2.54l.292-.159a.873.873 0 0 1 1.255.52l.094.319c.527 1.79 3.065 1.79 3.592 0l.094-.319a.873.873 0 0 1 1.255-.52l.292.16c1.64.893 3.434-.902 2.54-2.541l-.159-.292a.873.873 0 0 1 .52-1.255l.319-.094c1.79-.527 1.79-3.065 0-3.592l-.319-.094a.873.873 0 0 1-.52-1.255l.16-.292c.893-1.64-.902-3.433-2.541-2.54l-.292.159a.873.873 0 0 1-1.255-.52l-.094-.319zm-2.633.283c.246-.835 1.428-.835 1.674 0l.094.319a1.873 1.873 0 0 0 2.693 1.115l.291-.16c.764-.415 1.6.42 1.184 1.185l-.159.292a1.873 1.873 0 0 0 1.116 2.692l.318.094c.835.246.835 1.428 0 1.674l-.319.094a1.873 1.873 0 0 0-1.115 2.693l.16.291c.415.764-.42 1.6-1.185 1.184l-.291-.159a1.873 1.873 0 0 0-2.693 1.116l-.094.318c-.246.835-1.428.835-1.674 0l-.094-.319a1.873 1.873 0 0 0-2.692-1.115l-.292.16c-.764.415-1.6-.42-1.184-1.185l.159-.291A1.873 1.873 0 0 0 1.945 8.93l-.319-.094c-.835-.246-.835-1.428 0-1.674l.319-.094A1.873 1.873 0 0 0 3.06 4.377l-.16-.292c-.415-.764.42-1.6 1.185-1.184l.292.159a1.873 1.873 0 0 0 2.692-1.115l.094-.319z"></path></svg>')

    ui.popover(
        bs_gear_icon(icon_title),
        title = icon_title,
        ui.input_slider("n", "Number of points", 1, 100, 50)
    )
    ```

    ```python
    icon_title = "Settings"
    def fa_gear_icon(title: str):
        # Enhanced from https://rstudio.github.io/fontawesome/ via `fontawesome::fa("gear", a11y = "sem", title = icon_title)`
        return ui.HTML(f'<svg aria-label="{title}" role="img" viewBox="0 0 512 512" style="height:1em;width:1em;vertical-align:-0.125em;margin-left:auto;margin-right:auto;font-size:inherit;fill:currentColor;overflow:visible;position:relative;"><title>{title}</title><path d="M495.9 166.6c3.2 8.7 .5 18.4-6.4 24.6l-43.3 39.4c1.1 8.3 1.7 16.8 1.7 25.4s-.6 17.1-1.7 25.4l43.3 39.4c6.9 6.2 9.6 15.9 6.4 24.6c-4.4 11.9-9.7 23.3-15.8 34.3l-4.7 8.1c-6.6 11-14 21.4-22.1 31.2c-5.9 7.2-15.7 9.6-24.5 6.8l-55.7-17.7c-13.4 10.3-28.2 18.9-44 25.4l-12.5 57.1c-2 9.1-9 16.3-18.2 17.8c-13.8 2.3-28 3.5-42.5 3.5s-28.7-1.2-42.5-3.5c-9.2-1.5-16.2-8.7-18.2-17.8l-12.5-57.1c-15.8-6.5-30.6-15.1-44-25.4L83.1 425.9c-8.8 2.8-18.6 .3-24.5-6.8c-8.1-9.8-15.5-20.2-22.1-31.2l-4.7-8.1c-6.1-11-11.4-22.4-15.8-34.3c-3.2-8.7-.5-18.4 6.4-24.6l43.3-39.4C64.6 273.1 64 264.6 64 256s.6-17.1 1.7-25.4L22.4 191.2c-6.9-6.2-9.6-15.9-6.4-24.6c4.4-11.9 9.7-23.3 15.8-34.3l4.7-8.1c6.6-11 14-21.4 22.1-31.2c5.9-7.2 15.7-9.6 24.5-6.8l55.7 17.7c13.4-10.3 28.2-18.9 44-25.4l12.5-57.1c2-9.1 9-16.3 18.2-17.8C227.3 1.2 241.5 0 256 0s28.7 1.2 42.5 3.5c9.2 1.5 16.2 8.7 18.2 17.8l12.5 57.1c15.8 6.5 30.6 15.1 44 25.4l55.7-17.7c8.8-2.8 18.6-.3 24.5 6.8c8.1 9.8 15.5 20.2 22.1 31.2l4.7 8.1c6.1 11 11.4 22.4 15.8 34.3zM256 336a80 80 0 1 0 0-160 80 80 0 1 0 0 160z"/></svg>')
    ui.popover(
        fa_gear_icon(icon_title),
        title = icon_title,
        ui.input_slider("n", "Number of points", 1, 100, 50)
    )
    ```

    See Also
    --------
    * <https://getbootstrap.com/docs/5.3/components/popovers/>
    * :func:`~shiny.ui.update_popover`
    * :func:`~shiny.ui.tooltip`
    """

    # * If you're using [bsicons::bs_icon()], provide a `title`.
    # * If you're using [fontawesome::fa()], set `a11y = "sem"` and provide a `title`.

    # Theming/Styling
    # ---------------
    #
    # Like other bslib components, popovers can be themed by supplying [relevant theming
    # variables](https://rstudio.github.io/bslib/articles/bs5-variables.html#popover-bg)
    # to [bs_theme()], which effects styling of every popover on the page. To style a
    # _specific_ popover differently from other popovers, utilize the `customClass`
    # option:
    #
    # ```
    # popover(
    #     "Trigger", "Popover message",
    #     options = {"customClass": "my-pop"}
    # )
    # ```
    #
    # And then add relevant rules to [bs_theme()] via [bs_add_rules()]:
    #
    # ```
    # bs_theme() |> bs_add_rules(".my-pop { max-width: none; }")
    # ```

    attrs, children = consolidate_attrs(*args, **kwargs)
    if len(children) == 0:
        raise RuntimeError("At least one value must be provided to `popover(*args)`.")

    if options:
        for name in ("content", "title", "placement"):
            if name in options:
                raise RuntimeError(
                    f"The key `{name}` in `popover(options=)` cannot be specified directly."
                )

    res = web_component(
        "bslib-popover",
        # Use display:none instead of <template> since shiny.js
        # doesn't bind to the contents of the latter
        tags.template(
            div(*children, style="display:contents;"),
            div(title, style="display:contents;"),
        ),
        trigger,
        id=resolve_id_or_none(id),
        placement=placement,
        bsOptions=json.dumps(drop_none(options or {})),
        **attrs,
    )

    return res



from __future__ import annotations

# TODO-barret-render.data_frame; Docs
# TODO-barret-render.data_frame; Add examples of selection!
import warnings
from typing import Any, Literal, Set, Union, cast

from ..._deprecated import warn_deprecated
from ..._typing_extensions import TypedDict
from ...types import ListOrTuple
from ._types import DataFrame, FrameRenderSelectionModes

NoneSelectionMode = Literal["none"]
RowSelectionMode = Literal["row", "rows"]
ColSelectionMode = Literal["col", "cols"]
RectSelectionMode = Literal["cell", "region"]

SelectionMode = Union[
    RowSelectionMode,
    ColSelectionMode,
    RectSelectionMode,
    NoneSelectionMode,
]
none_set: set[NoneSelectionMode] = {"none"}
row_set: set[RowSelectionMode] = {"row", "rows"}
col_set: set[ColSelectionMode] = {"col", "cols"}
rect_set: set[RectSelectionMode] = {"cell", "region"}
complete_selection_mode_set: set[SelectionMode] = (
    none_set | row_set | col_set | rect_set
)


# Types
class SelectionModes:

    row: Literal["none", "single", "multiple"]
    col: Literal["none", "single", "multiple"]
    rect: Literal["none", "cell", "region"]

    def as_dict(self) -> FrameRenderSelectionModes:
        return {
            "row": self.row,
            "col": self.col,
            "rect": self.rect,
        }

    def __init__(self, *, selection_mode_set: set[SelectionMode]):
        self.row = "none"
        self.col = "none"
        self.rect = "none"

        if not selection_mode_set.issubset(complete_selection_mode_set):
            bad_modes = selection_mode_set - complete_selection_mode_set
            # TODO-test; Test that this error is raised
            raise ValueError(
                f"Unknown selection modes: {', '.join(bad_modes)}. "
                f"Valid selection modes: {', '.join(complete_selection_mode_set)}"
            )

        if len(selection_mode_set) == 0:
            selection_mode_set = {"none"}
        elif "none" in selection_mode_set and len(selection_mode_set) > 1:
            extra_set = selection_mode_set - none_set
            # TODO-test; Test that this error is raised
            raise ValueError(
                "Cannot have other selection modes included with `none`. "
                f"Extra selection modes: {', '.join(extra_set)}"
            )

        if "row" in selection_mode_set and "rows" in selection_mode_set:
            raise ValueError("Cannot have both `row` and `rows` in selection modes.")

        if "col" in selection_mode_set and "cols" in selection_mode_set:
            raise ValueError("Cannot have both `col` and `cols` in selection modes.")

        if "cell" in selection_mode_set and "region" in selection_mode_set:
            raise ValueError("Cannot have both `cell` and `region` in selection modes.")

        if "row" in selection_mode_set:
            self.row = "single"
        elif "rows" in selection_mode_set:
            self.row = "multiple"

        if "col" in selection_mode_set:
            raise RuntimeError(
                "Column based cell selections are not currently supported."
            )
            self.col = "single"
        elif "cols" in selection_mode_set:
            raise RuntimeError(
                "Column based cell selections are not currently supported."
            )
            self.col = "multiple"

        if "cell" in selection_mode_set:
            raise RuntimeError(
                "Cell based cell selections are not currently supported."
            )
            self.rect = "cell"
        elif "region" in selection_mode_set:
            raise RuntimeError(
                "Region based cell selections are not currently supported."
            )
            self.rect = "region"

    def _is_none(self) -> bool:
        """
        Check if all selection modes are set to `"none"`.
        """
        return (self.row == "none") and (self.col == "none") and (self.rect == "none")

    def _has_row(self) -> bool:
        """
        Check if `.row` does not equal `"none"`.
        """
        return self.row != "none"

    def _has_col(self) -> bool:
        """
        Check if `.col` does not equal `"none"`.
        """
        return self.col != "none"

    def _has_rect(self) -> bool:
        """
        Check if `.rect` does not equal `"none"`.
        """
        return self.rect != "none"


SelectionModeInput = Union[
    # String
    SelectionMode,
    # Sequence of strings
    ListOrTuple[SelectionMode],
    # Set[SelectionMode],
    # Support the upgraded selection modes
    SelectionModes,
]
"""
Input types for selection modes in a DataGrid or DataTable.
"""


# ####################
# # BrowserCellSelection: For receiving selection info from JS
# ####################

# Should only contain a single selection area

# Do not include `BrowserCellSelectionAll` as it should be represented by a row, column, or region with appropriate values.
# # class BrowserCellSelectionAll(TypedDict):
# #    type: Literal["all"]


class BrowserCellSelectionNone(TypedDict):
    type: Literal["none"]


class BrowserCellSelectionRow(TypedDict):
    type: Literal["row"]
    rows: ListOrTuple[int]


class BrowserCellSelectionCol(TypedDict):
    type: Literal["col"]
    cols: ListOrTuple[int]


class BrowserCellSelectionRect(TypedDict):
    type: Literal["rect"]
    # All rows and cols values! (Not min/max values!)
    # This is required as row selection is applied after sort/filter is applied. This
    # rearranges the rows into a non-sequential order, requiring all row numbers.
    rows: ListOrTuple[int]
    cols: ListOrTuple[int]


# For receiving selection info from JS:
BrowserCellSelection = Union[
    BrowserCellSelectionRow,
    BrowserCellSelectionCol,
    BrowserCellSelectionRect,
    BrowserCellSelectionNone,
]
"""
A single data frame selection set sent to/from the browser.

This object will be converted to `CellSelection` when being sent to (or received from)
the browser.

`type` values:

- `"none"`: No cells are selected. `rows` and `cols` will be missing.
- `"row"`: A set of selected `rows` numbers. `cols` will be missing.
- `"col"`: A set of selected `cols` numbers. `rows` will be missing.
- `"rect"`: A single rectangular region that is selected. `rows` and `cols` will be the
  row and column numbers for the selected region.
"""


class CellSelection(TypedDict):
    """
    A single data frame selection enhanced with missing keys.

    This object will always contain `rows` and `cols` keys to make it more user friendly.

    * If a cell selection is being sent to the browser, unnecessary keys will be dropped.
    * If a cell selection is being received from the browser, missing `rows` and `cols`
      keys will be added to contain all possible values. E.g. when `type="row"`, `cols`
      will be set to all column numbers for the data. These _extra_ values are not sent
      across as an input as they are not needed for the browser to render the selection.

    `type` values:

    - `"none"`: No cells are selected. `rows` and `cols` will be empty tuples.
    - `"row"`: A set of selected `rows` numbers. `cols` will be all column numbers for the data.
    - `"col"`: A set of selected `cols` numbers. `rows` will be all row numbers for the data.
    - `"rect"`: A single rectangular region that is selected. `rows` and `cols` will be
      the row and column numbers for the selected region.
    """

    type: Literal["none", "row", "col", "rect"]
    rows: ListOrTuple[int]
    cols: ListOrTuple[int]


# ####################


def as_browser_cell_selection(
    x: BrowserCellSelection | CellSelection | Literal["all"] | None,
    *,
    selection_modes: SelectionModes,
    nw_data: DataFrame[Any],
) -> BrowserCellSelection:

    if x is None or selection_modes._is_none():
        return {"type": "none"}

    if x == "all":
        row_len, col_len = nw_data.shape
        # Look at the selection modes to determine what to do
        if selection_modes._has_rect():
            if selection_modes.rect == "cell":
                warnings.warn(
                    "Cannot select all cells with `cell` selection mode. Selecting the first cell",
                    stacklevel=3,
                )
                return {"type": "rect", "rows": (0,), "cols": (0,)}
            if selection_modes.rect == "region":
                return {
                    "type": "rect",
                    "rows": tuple(range(row_len)),
                    "cols": tuple(range(col_len)),
                }
        if selection_modes._has_row():
            if selection_modes.row == "single":
                warnings.warn(
                    "Cannot select all rows with `row` selection mode. Selecting the first row",
                    stacklevel=3,
                )
                return {"type": "row", "rows": (0,)}
            if selection_modes.row == "multiple":
                return {"type": "row", "rows": tuple(range(row_len))}
        if selection_modes._has_col():
            if selection_modes.col == "single":
                warnings.warn(
                    "Cannot select all columns with `col` selection mode. Selecting the first column",
                    stacklevel=3,
                )
                return {"type": "col", "cols": (0,)}
            if selection_modes.col == "multiple":
                return {"type": "col", "cols": tuple(range(col_len))}
        raise ValueError(
            "Current selection modes do not support cell based selection. "
            f"Current selection modes: {selection_modes}"
        )

    # Can't handle lists of dictionaries right now
    assert isinstance(x, dict)

    def to_int_tuple_or_none(
        arr: int | ListOrTuple[int] | None, *, name: str
    ) -> tuple[int, ...] | None:
        if arr is None:
            return None
        if not isinstance(arr, (list, tuple)):
            arr = (arr,)

        for item in arr:
            if not isinstance(item, int):
                raise TypeError(
                    f"Expected cell selection's `{name}` to be an int. Received {type(item)}"
                )
        return tuple(arr)

    rows = to_int_tuple_or_none(x.get("rows", None), name="rows")
    cols = to_int_tuple_or_none(x.get("cols", None), name="cols")

    assert "type" in x, "`type` field is required in CellSelection"

    if x["type"] == "none":
        return {"type": "none"}

    if x["type"] == "row":
        if not selection_modes._has_row():
            raise ValueError(
                "Current selection modes do not support row based selection. "
                f"Current selection modes: {selection_modes}"
            )
        assert rows is not None
        return {"type": "row", "rows": rows}
    if x["type"] == "col":
        assert cols is not None
        return {"type": "col", "cols": cols}
    if x["type"] == "rect":
        if not selection_modes._has_rect():
            raise ValueError(
                "Current selection modes do not support cell based selection. "
                f"Current selection modes: {selection_modes}"
            )
        assert rows is not None
        assert cols is not None
        assert len(rows) > 0
        assert len(cols) > 0
        return {"type": "rect", "rows": rows, "cols": cols}

    raise ValueError(f"Unhandled CellSelection['type']: {x['type']}")


def as_cell_selection(
    x: CellSelection | Literal["all"] | None | BrowserCellSelection,
    *,
    selection_modes: SelectionModes,
    nw_data: DataFrame[Any],
    data_view_rows: ListOrTuple[int],
    data_view_cols: ListOrTuple[int],
) -> CellSelection:
    """
    Converts the selection to `BrowserCellSelection` type and then adds missing
    `rows` and `cols` to represent all reasonable values for consistent user interactions.
    """
    browser_cell_selection = as_browser_cell_selection(
        x,
        selection_modes=selection_modes,
        nw_data=nw_data,
    )
    ret: CellSelection | None = None
    if browser_cell_selection["type"] == "none":
        ret = {
            "type": "none",
            "rows": (),
            "cols": (),
        }
    elif browser_cell_selection["type"] == "row":

        ret = {
            "type": "row",
            "rows": browser_cell_selection["rows"],
            "cols": tuple(data_view_cols),
        }
    elif browser_cell_selection["type"] == "col":
        ret = {
            "type": "col",
            "rows": tuple(data_view_rows),
            "cols": browser_cell_selection["cols"],
        }
    elif browser_cell_selection["type"] == "rect":
        ret = {
            "type": "rect",
            "rows": browser_cell_selection["rows"],
            "cols": browser_cell_selection["cols"],
        }
    else:
        raise ValueError(
            f"Unhandled BrowserCellSelection['type']: {browser_cell_selection['type']}"
        )

    # Make sure the rows are within the data
    nrow, ncol = nw_data.shape
    ret["rows"] = tuple(row for row in ret["rows"] if row < nrow)
    ret["cols"] = tuple(col for col in ret["cols"] if col < ncol)

    return ret


RowSelectionModeDeprecated = Literal["single", "multiple", "none", "deprecated"]


def as_selection_modes(
    selection_mode: SelectionModeInput,
    *,
    name: str,
    row_selection_mode: RowSelectionModeDeprecated = "deprecated",
) -> SelectionModes:
    # TODO-barret; Test that as _selection_modes can take and return a SelectionModes object
    if isinstance(selection_mode, SelectionModes):
        return selection_mode

    selection_mode_val = selection_mode

    # If a user supplied `row_selection_mode`, use the value but warn against it
    if row_selection_mode != "deprecated":

        if row_selection_mode == "none":
            selection_mode_val = "none"
        elif row_selection_mode == "single":
            selection_mode_val = "row"
        elif row_selection_mode == "multiple":
            selection_mode_val = "rows"
        else:
            raise ValueError(f"Unknown row_selection_mode: {row_selection_mode}")

        warn_deprecated(
            f"`{name}(row_selection_mode=)` has been superseded by `{name}(selection_mode=)`."
            f' Please use `{name}(selection_mode="{selection_mode_val}")` instead.'
        )
    if not isinstance(selection_mode_val, tuple):
        selection_mode_val = (selection_mode_val,)

    selection_mode_set = cast(Set[SelectionMode], set(selection_mode_val))

    return SelectionModes(selection_mode_set=selection_mode_set)



from faicons import icon_svg

from shiny import App, Inputs, ui

app_ui = ui.page_fluid(
    ui.h1("Accordion Kitchensink"),
    ui.accordion(
        ui.accordion_panel(
            "Panel 1",
            ui.p("This is the content of Panel 1"),
            value="panel1",
        ),
        ui.accordion_panel(
            "Panel 2",
            ui.p("This is the content of Panel 2"),
            icon=icon_svg("trash-arrow-up"),
            value="panel2",
        ),
        id="accordion_1",
        width="600px",
        height="300px",
        multiple=False,
        class_="bg-light",
    ),
    ui.accordion(
        ui.accordion_panel(
            "Panel 3",
            ui.p("This is the content of Panel 3"),
            value="panel3",
        ),
        ui.accordion_panel(
            "Panel 4",
            ui.p("This is the content of Panel 4"),
            value="panel4",
        ),
        id="accordion_2",
        multiple=True,
    ),
)


def server(input: Inputs):
    pass


app = App(app_ui, server)
